# -*- coding: utf-8 -*
"""Convert ISA-JSON to SRA-XML"""
import logging

from isatools import isajson, sra


log = logging.getLogger('isatools')


def convert(json_fp, path, config_dir=None, sra_settings=None,
            datafilehashes=None, validate_first=True):
    """ Converter for ISA-JSON to SRA.
    :param json_fp: File pointer to ISA JSON input
    :param path: Directory for output SRA XMLs to be written
    :param config_dir: path to JSON configuration. If none, uses default
        embedded in API
    :param sra_settings: SRA settings dict
    :param datafilehashes: Data files with hashes, in a dict
    """
    if validate_first:
        log.info("Validating input JSON before conversion")
        report = isajson.validate(fp=json_fp, config_dir=config_dir,
                                  log_level=logging.ERROR)
        if len(report.get('errors')) > 0:
            log.fatal("Could not proceed with conversion as there are some "
                      "validation errors. Check log.")
            return
    log.info("Loading isajson {}".format(json_fp.name))
    isa = isajson.load(fp=json_fp)
    log.info("Exporting SRA to {}".format(path))
    log.debug("Using SRA settings ".format(sra_settings))
    sra.export(isa, path, sra_settings=sra_settings,
               datafilehashes=datafilehashes)


"""
sra_settings = {
 "sra_center": “EI",
  "sra_broker": “EI",
  "sra_action": “ADD”,
 “sra_broker_inform_on_status”: “support@copo.org”,
 “sra_broker_inform_on_error”: “support@copo.org"
}
datafilehashes = {
   "myfile1.fastq": "3a7886617efd0c8f76c360e944149462",
   "myfile2.fastq": "9918006f1eeff68e695539c8843df334"
}
json2sra.convert(json_fp, path, sra_settings=sra_settings,
filehashes=datafilehashes)

If files in filehashes dict don't map 1:1 to files found in ISA JSON content,
raise Exception

json2sra.convert(json_fp=open('/Users/dj/PycharmProjects/isa-api/copo.json'),
path='/Users/dj/PycharmProjects/isa-api/tmp', sra_settings=)
"""



# -*- coding: utf-8 -*-
"""Functions for retrieving metadata from MetaboLights.

This module connects to the European Bioinformatics Institute's
MetaboLights database. If you have problems with it, check that
it's working at http://www.ebi.ac.uk/metabolights/
"""
from __future__ import absolute_import
import ftplib
import glob
import json
import logging
import os
import re
import shutil
import tempfile

import pandas as pd
# import modin.pandas as pd_modin

from isatools import isatab
from isatools.convert import isatab2json
from isatools.model import OntologyAnnotation


EBI_FTP_SERVER = 'ftp.ebi.ac.uk'
MTBLS_BASE_DIR = '/pub/databases/metabolights/studies/public'


log = logging.getLogger('isatools')

# REGEXES
_RX_FACTOR_VALUE = re.compile(r'Factor Value\[(.*?)\]')


def get(mtbls_study_id, target_dir=None):
    """
    This function downloads ISA content from the MetaboLights FTP site.

    :param mtbls_study_id: Study identifier for MetaboLights study to get, as
    a str (e.g. MTBLS1)
    :param target_dir: Path to write files to. If None, writes to temporary
    directory (generated on the fly)
    :return: Path where the files were written to

    Example usage:
        isa_json = MTBLS.get_study('MTBLS1', '/tmp/mtbls')
    """
    log.info('Setting up ftp with {}'.format(EBI_FTP_SERVER))
    ftp = ftplib.FTP(EBI_FTP_SERVER)

    log.info('Logging in as anonymous user...')
    response = ftp.login()

    if '230' in response:  # 230 means Login successful
        log.info('Log in successful!')

        try:
            log.info("Looking for study '{}'".format(mtbls_study_id))
            ftp.cwd('{base_dir}/{study}'.format(
                base_dir=MTBLS_BASE_DIR, study=mtbls_study_id))

            if target_dir is None:
                target_dir = tempfile.mkdtemp()
            log.info("Using directory '{}'".format(target_dir))

            files = list(ftp.nlst())
            investigation_filename = None
            try:
                investigation_filename = next(
                    filter(lambda x: x.startswith('i_') and
                           x.endswith('.txt'), files))
            except StopIteration:
                log.fatal(
                    'Could not find an investigation file for this study')

            if investigation_filename is not None:
                with open(os.path.join(
                        target_dir, investigation_filename), 'wb') as out_file:
                    log.info(
                        "Retrieving file '{}'".format(
                            EBI_FTP_SERVER + MTBLS_BASE_DIR + '/' +
                            mtbls_study_id + '/' + investigation_filename))
                    ftp.retrbinary('RETR {i_file}'.format(
                        i_file=investigation_filename), out_file.write)

                with open(out_file.name, encoding='utf-8') as i_fp:
                    i_bytes = i_fp.read()

                    lines = i_bytes.splitlines()
                    s_filenames = [l.split('\t')[1][1:-1] for l in lines if
                                   'Study File Name' in l]

                    for s_filename in s_filenames:
                        with open(os.path.join(target_dir, s_filename), 'wb') \
                                as out_file:
                            log.info("Retrieving file '{}'".format(
                                EBI_FTP_SERVER + MTBLS_BASE_DIR + '/' +
                                mtbls_study_id + '/' + s_filename))
                            ftp.retrbinary(
                                'RETR {s_file}'.format(
                                    s_file=s_filename), out_file.write)

                    a_filenames_lines = [l.split('\t') for l in lines if
                                         'Study Assay File Name' in l]
                    for a_filename_line in a_filenames_lines:
                        for a_filename in [f[1:-1] for f in
                                           a_filename_line[1:]]:
                            with open(os.path.join(target_dir, a_filename),
                                      'wb') as out_file:
                                log.info("Retrieving file '{}'".format(
                                    EBI_FTP_SERVER + MTBLS_BASE_DIR + '/' +
                                    mtbls_study_id + '/' + a_filename))
                                ftp.retrbinary(
                                    'RETR {a_file}'.format(
                                        a_file=a_filename), out_file.write)

        except ftplib.error_perm as ftperr:
            log.fatal(
                "Could not retrieve MetaboLights study '{study}': {error}"
                .format(study=mtbls_study_id, error=ftperr))
        finally:
            ftp.close()
            return target_dir
    else:
        ftp.close()
        raise ConnectionError(
            'There was a problem connecting to MetaboLights: {response}'
            .format(response=response))


def getj(mtbls_study_id):
    """
    This function downloads the specified MetaboLights study and returns an
    ISA JSON representation of it

    :param mtbls_study_id: Study identifier for MetaboLights study to get,
    as a str (e.g. MTBLS1)
    :return: ISA JSON representation of the MetaboLights study

    Example usage:
        isa_json = MTBLS.load('MTBLS1')
    """
    tmp_dir = get(mtbls_study_id)
    if tmp_dir is None:
        raise IOError('There was a problem retrieving the study {study_id}'
                      .format(study_id=mtbls_study_id))

    isa_json = isatab2json.convert(
        tmp_dir, identifier_type=isatab2json.IdentifierType.name,
        validate_first=False, use_new_parser=True)

    shutil.rmtree(tmp_dir)
    return isa_json


def get_data_files(mtbls_study_id, factor_selection=None):
    tmp_dir = get(mtbls_study_id)
    if tmp_dir is None:
        raise IOError(
            'There was a problem retrieving study {study_id}. Does it exist?'
            .format(study_id=mtbls_study_id))

    else:
        result = slice_data_files(tmp_dir, factor_selection=factor_selection)

    shutil.rmtree(tmp_dir)
    return result


def slice_data_files(dir, factor_selection=None):
    """
    This function gets a list of samples and related data file URLs for a given
    MetaboLights study, optionally filtered by factor value (currently by
    matching on exactly 1 factor value)

    :param mtbls_study_id: Study identifier for MetaboLights study to get, as
    a str (e.g. MTBLS1)
    :param factor_selection: A list of selected factor values to filter on
    samples
    :return: A list of dicts {sample_name, list of data_files} containing
    sample names with associated data filenames

    Example usage:
        samples_and_data = mtbls.get_data_files('MTBLS1', [{'Gender': 'Male'}])

    TODO:  Need to work on more complex filters e.g.:
        {"gender": ["male", "female"]} selects samples matching "male" or
        "female" factor value
        {"age": {"equals": 60}} selects samples matching age 60
        {"age": {"less_than": 60}} selects samples matching age less than 60
        {"age": {"more_than": 60}} selects samples matching age more than 60

        To select samples matching "male" and age less than 60:
        {
            "gender": "male",
            "age": {
                "less_than": 60
            }
        }
    """
    results = []
    # first collect matching samples
    for table_file in glob.iglob(os.path.join(dir, '[a|s]_*')):
        log.info('Loading {table_file}'.format(table_file=table_file))

        with open(table_file, encoding='utf-8') as fp:
            df = isatab.load_table(fp)
            df = df[[x for x in df.columns if
                     'Factor Value' in x or 'Sample Name' in x]]
            df.columns = ['sample' if 'Sample Name' in x else x for x in
                          df.columns]
            df.columns = [x[13:-1] if 'Factor Value' in x else x for x in
                          df.columns]
            df.columns = [x.replace(' ', '_') for x in df.columns]
            # build query
            sample_names_series = df['sample'].drop_duplicates()
            if factor_selection is None:
                results = sample_names_series.apply(lambda x: {
                                'sample': x,
                                'data_files': [],
                                'query_used': ''
                            }).tolist()
            else:
                factor_query = ''
                for factor_name, factor_value in factor_selection.items():
                    factor_name = factor_name.replace(' ', '_')
                    factor_query += '{factor_name}=="{factor_value}" and '\
                        .format(factor_name=factor_name,
                                factor_value=factor_value)
                factor_query = factor_query[:-5]
                try:
                    query_results = df.query(factor_query)[
                        'sample'].drop_duplicates()
                    results = query_results.apply(lambda x: {
                        'sample': x,
                        'data_files': [],
                        'query_used': factor_selection
                    }).tolist()
                except pd.core.computation.ops.UndefinedVariableError:
                    pass

    # now collect the data files relating to the samples
    for table_file in glob.iglob(os.path.join(dir, 'a_*.txt')):
        with open(table_file, encoding='utf-8') as fp:
            df = isatab.load_table(fp)
            df = df[[x for x in df.columns if
                     'File' in x or 'Sample Name' in x]]
            df.columns = ['sample' if 'Sample Name' in x else x for x in
                          df.columns]
            for result in results:
                sample_name = result['sample']
                sample_rows = df.loc[df['sample'] == sample_name]

                for data_col in [x for x in sample_rows.columns
                                 if 'File' in x]:
                    data_files = sample_rows[data_col]
                    result['data_files'] = [i for i in data_files if
                                            str(i) != 'nan']
    return results


def get_factor_names(mtbls_study_id):
    """
    This function gets the factor names used in a MetaboLights study

    :param mtbls_study_id: Accession number of the MetaboLights study
    :return: A set of factor names used in the study

    Example usage:
        factor_names = get_factor_names('MTBLS1')
    """
    tmp_dir = get(mtbls_study_id)

    factors = set()

    for table_file in glob.iglob(os.path.join(tmp_dir, '[a|s]_*')):
        with open(os.path.join(tmp_dir, table_file), encoding='utf-8') as fp:
            df = isatab.load_table(fp)

            factors_headers = [header for header in list(df.columns.values)
                               if _RX_FACTOR_VALUE.match(header)]

            for header in factors_headers:
                factors.add(header[13:-1])
    return factors


def get_factor_values(mtbls_study_id, factor_name):
    """
    This function gets the factor values of a factor in a MetaboLights study

    :param mtbls_study_id: Accession number of the MetaboLights study
    :param factor_name: The factor name for which values are being queried
    :return: A set of factor values associated with the factor and study

    Example usage:
        factor_values = get_factor_values('MTBLS1', 'genotype')
    """
    tmp_dir = get(mtbls_study_id)

    fvs = set()

    for table_file in glob.iglob(os.path.join(tmp_dir, '[a|s]_*')):
        with open(os.path.join(tmp_dir, table_file), encoding='utf-8') as fp:
            df = isatab.load_table(fp)

            if 'Factor Value[{factor}]'.format(factor=factor_name) in \
                    list(df.columns.values):
                for _, match in df[
                    'Factor Value[{factor}]'.format(
                        factor=factor_name)].iteritems():
                    try:
                        match = match.item()
                    except AttributeError:
                        pass

                    if isinstance(match, (str, int, float)):
                        if str(match) != 'nan':
                            fvs.add(match)
    shutil.rmtree(tmp_dir)
    return fvs


def load(mtbls_study_id):
    tmp_dir = get(mtbls_study_id)

    if tmp_dir is None:
        raise IOError(
            'There was a problem retrieving the study {study_id}'.format(
                study_id=mtbls_study_id))

    with open(glob.glob(os.path.join(tmp_dir, 'i_*.txt'))[0],
              encoding='utf-8') as f:
        ISA = isatab.load(f)

        shutil.rmtree(tmp_dir)
        return ISA


def get_factors_summary(mtbls_study_id):
    """
    This function generates a factors summary for a MetaboLights study

    :param mtbls_study_id: Accession number of the MetaboLights study
    :return: A list of dicts summarising the set of factor names and values
    associated with each sample

    Note: it only returns a summary of factors with variable values.

    Example usage:
        factor_summary = get_factors_summary('MTBLS1')
        [
            {
                "name": "ADG19007u_357",
                "Metabolic syndrome": "Control Group",
                "Gender": "Female"
            },
            {
                "name": "ADG10003u_162",
                "Metabolic syndrome": "diabetes mellitus",
                "Gender": "Female"
            },
        ]
    """
    ISA = load(mtbls_study_id=mtbls_study_id)

    all_samples = []
    for study in ISA.studies:
        all_samples.extend(study.samples)

    samples_and_fvs = []

    for sample in all_samples:
        sample_and_fvs = {
                'sources': ';'.join([x.name for x in sample.derives_from]),
                'sample': sample.name,
            }

        for fv in sample.factor_values:
            if isinstance(fv.value, (str, int, float)):
                fv_value = fv.value
            elif isinstance(fv.value, OntologyAnnotation):
                fv_value = fv.value.term

            sample_and_fvs[fv.factor_name.name] = fv_value

        samples_and_fvs.append(sample_and_fvs)

    df = pd.DataFrame(samples_and_fvs)
    nunique = df.apply(pd.Series.nunique)
    cols_to_drop = nunique[nunique == 1].index

    df = df.drop(cols_to_drop, axis=1)
    return df.to_dict(orient='records')


def get_study_groups(mtbls_study_id):
    factors_summary = get_factors_summary(mtbls_study_id=mtbls_study_id)
    study_groups = {}

    for factors_item in factors_summary:
        fvs = tuple(
            factors_item[k] for k in factors_item.keys() if k != 'name')

        if fvs in study_groups.keys():
            study_groups[fvs].append(factors_item['name'])
        else:
            study_groups[fvs] = [factors_item['name']]
    return study_groups


def get_study_groups_samples_sizes(mtbls_study_id):
    study_groups = get_study_groups(mtbls_study_id=mtbls_study_id)
    return list(map(lambda x: (x[0], len(x[1])), study_groups.items()))


def get_sources_for_sample(mtbls_study_id, sample_name):
    ISA = load(mtbls_study_id=mtbls_study_id)
    hits = []

    for study in ISA.studies:
        for sample in study.samples:
            if sample.name == sample_name:
                print('found a hit: {sample_name}'.format(
                    sample_name=sample.name))

                for source in sample.derives_from:
                    hits.append(source.name)
    return hits


def get_data_for_sample(mtbls_study_id, sample_name):
    ISA = load(mtbls_study_id=mtbls_study_id)
    hits = []
    for study in ISA.studies:
        for assay in study.assays:
            for data in assay.data_files:
                if sample_name in [x.name for x in data.generated_from]:
                    log.info('found a hit: {filename}'.format(
                        filename=data.filename))
                    hits.append(data)
    return hits


def get_study_groups_data_sizes(mtbls_study_id):
    study_groups = get_study_groups(mtbls_study_id=mtbls_study_id)
    return list(map(lambda x: (x[0], len(x[1])), study_groups.items()))


def get_characteristics_summary(mtbls_study_id):
    """
        This function generates a characteristics summary for a MetaboLights
        study

        :param mtbls_study_id: Accession number of the MetaboLights study
        :return: A list of dicts summarising the set of characteristic names
        and values associated with each sample

        Note: it only returns a summary of characteristics with variable
        values.

        Example usage:
            characteristics_summary = get_characteristics_summary('MTBLS5')
            [
                {
                    "name": "6089if_9",
                    "Variant": "Synechocystis sp. PCC 6803.sll0171.ko"
                },
                {
                    "name": "6089if_43",
                    "Variant": "Synechocystis sp. PCC 6803.WT.none"
                },
            ]


        """
    ISA = load(mtbls_study_id=mtbls_study_id)

    all_samples = []
    for study in ISA.studies:
        all_samples.extend(study.samples)

    samples_and_characs = []
    for sample in all_samples:
        sample_and_characs = {
                'name': sample.name
            }

        for source in sample.derives_from:
            for c in source.characteristics:
                if isinstance(c.value, (str, int, float)):
                    c_value = c.value
                elif isinstance(c.value, OntologyAnnotation):
                    c_value = c.value.term
                sample_and_characs[c.category.term] = c_value

        samples_and_characs.append(sample_and_characs)

    df = pd.DataFrame(samples_and_characs)
    nunique = df.apply(pd.Series.nunique)
    cols_to_drop = nunique[nunique == 1].index

    df = df.drop(cols_to_drop, axis=1)
    return df.to_dict(orient='records')


def get_study_variable_summary(mtbls_study_id):
    ISA = load(mtbls_study_id=mtbls_study_id)

    all_samples = []
    for study in ISA.studies:
        all_samples.extend(study.samples)

    samples_and_variables = []
    for sample in all_samples:
        sample_and_vars = {
            'sample_name': sample.name
        }

        for fv in sample.factor_values:
            if isinstance(fv.value, (str, int, float)):
                fv_value = fv.value
            elif isinstance(fv.value, OntologyAnnotation):
                fv_value = fv.value.term
            sample_and_vars[fv.factor_name.name] = fv_value

        for source in sample.derives_from:
            sample_and_vars['source_name'] = source.name
            for c in source.characteristics:
                if isinstance(c.value, (str, int, float)):
                    c_value = c.value
                elif isinstance(c.value, OntologyAnnotation):
                    c_value = c.value.term
                sample_and_vars[c.category.term] = c_value

        samples_and_variables.append(sample_and_vars)

    df = pd.DataFrame(samples_and_variables)
    nunique = df.apply(pd.Series.nunique)
    cols_to_drop = nunique[nunique == 1].index

    df = df.drop(cols_to_drop, axis=1)
    return df.to_dict(orient='records')


def get_study_group_factors(mtbls_study_id):
    factors_list = []
    tmp_dir = get(mtbls_study_id)
    if tmp_dir is None:
        raise FileNotFoundError(
            'Could not download {study_id}'.format(study_id=mtbls_study_id))

    for table_file in glob.iglob(os.path.join(tmp_dir, '[a|s]_*')):
        with open(os.path.join(tmp_dir, table_file), encoding='utf-8') as fp:
            df = isatab.load_table(fp)

            factor_columns = [x for x in df.columns if x.startswith(
                'Factor Value')]
            if len(factor_columns) > 0:
                factors_list = df[factor_columns].drop_duplicates()\
                    .to_dict(orient='records')
    return factors_list


def get_filtered_df_on_factors_list(mtbls_study_id):
    factors_list = get_study_group_factors(mtbls_study_id=mtbls_study_id)
    queries = []

    for item in factors_list:
        query_str = []

        for k, v in item.items():
            k = k.replace(' ', '_').replace('[', '_').replace(']', '_')
            if isinstance(v, str):
                v = v.replace(' ', '_').replace('[', '_').replace(']', '_')
                query_str.append("{k} == '{v}' and ".format(k=k, v=v))

        query_str = ''.join(query_str)[:-4]
        queries.append(query_str)

    tmp_dir = get(mtbls_study_id)
    for table_file in glob.iglob(os.path.join(tmp_dir, '[a|s]_*')):
        with open(os.path.join(tmp_dir, table_file), encoding='utf-8') as fp:
            df = isatab.load_table(fp)

            cols = df.columns
            cols = cols.map(
                lambda x: x.replace(' ', '_') if isinstance(x, str) else x)
            df.columns = cols

            cols = df.columns
            cols = cols.map(
                lambda x: x.replace('[', '_') if isinstance(x, str) else x)
            df.columns = cols

            cols = df.columns
            cols = cols.map(
                lambda x: x.replace(']', '_') if isinstance(x, str) else x)
            df.columns = cols

        for query in queries:
            df2 = df.query(query)  # query uses pandas.eval, which evaluates
            # queries like pure Python notation
            if 'Sample_Name' in df.columns:
                print('Group: {query} / Sample_Name: {sample_name}'.format(
                    query=query, sample_name=list(df2['Sample_Name'])))

            if 'Source_Name' in df.columns:
                print('Group: {} / Sources_Name: {}'.format(
                    query, list(df2['Source_Name'])))

            if 'Raw_Spectral_Data_File' in df.columns:
                print('Group: {query} / Raw_Spectral_Data_File: {filename}'
                      .format(query=query[13:-2],
                              filename=list(df2['Raw_Spectral_Data_File'])))
    return queries


def get_mtbls_list():
    log.info(
        'Setting up ftp with {ftp_server}'.format(ftp_server=EBI_FTP_SERVER))
    ftp = ftplib.FTP(EBI_FTP_SERVER)

    log.info('Logging in as anonymous user...')
    response = ftp.login()

    mtbls_list = []
    if '230' in response:  # 230 means Login successful
        log.info('Log in successful!')

        try:
            ftp.cwd('{base_dir}'.format(base_dir=MTBLS_BASE_DIR))
            mtbls_list = ftp.nlst()
        except ftplib.error_perm as ftperr:
            log.error(
                'Could not get MTBLS directory list. Error: {err}'
                .format(err=ftperr))
    return mtbls_list


def dl_all_mtbls_isatab(target_dir):
    download_count = 0

    for i, mtblsid in enumerate(get_mtbls_list()):
        target_mtbls_subdir = os.path.join(target_dir, mtblsid)

        if not os.path.exists(target_mtbls_subdir):
            os.makedirs(target_mtbls_subdir)

        get(mtblsid, target_mtbls_subdir)
        download_count = i

    print('Downloaded {count} ISA-Tab studies from MetaboLights'.format(
        count=download_count))

# mtblisa commands


def get_study_command(isa_format, study_id, output):
    if os.path.exists(output):
        raise RuntimeError("Selected output path {} already exists!".format(
            output))

    if isa_format == "isa-tab":
        tmp_data = None
        try:
            log.info("Downloading study %s", study_id)
            tmp_data = get(study_id)
            if tmp_data is None:
                raise RuntimeError("Error downloading ISA study")

            log.debug(
                "Finished downloading data. Moving to final location %s",
                output)
            shutil.move(tmp_data, output)
            log.info("ISA archive written to %s", output)
        finally:
            if tmp_data:
                # try to clean up any temporary files left behind
                log.debug("Deleting %s, if there's anything there", tmp_data)
                shutil.rmtree(tmp_data, ignore_errors=True)
    elif isa_format == "isa-json":
        isajson = getj(study_id)
        if isajson is None:
            raise RuntimeError("Error downloading ISA study")
        log.debug(
            "Finished downloading data. Dumping json to final location %s",
            output)
        os.makedirs(output)
        json_file = os.path.join(output, "{}.json".format(
            isajson['identifier']))
        with open(json_file, 'w') as fd:
            json.dump(isajson, fd)
        log.info("ISA-JSON written to %s", output)
    else:
        raise ValueError("BUG! Got an invalid isa format '{}'".format(
            isa_format))


def get_factors_command(study_id, output):
    log.info("Getting factors for study %s. Writing to %s.",
             study_id, output.name)
    factor_names = get_factor_names(study_id)
    if factor_names is not None:
        json.dump(list(factor_names), output, indent=4)
        log.debug("Factor names written")
    else:
        raise RuntimeError("Error downloading factors.")


def get_factor_values_command(study_id, factor, output):
    log.info("Getting values for factor {factor} in study {study_id}. "
             "Writing to {output_file}.".format(
                  factor=factor, study_id=study_id, output_file=output.name))
    fvs = get_factor_values(study_id, factor)
    if fvs is not None:
        json.dump(list(fvs), output, indent=4)
        log.debug("Factor values written to {}".format(output))
    else:
        raise RuntimeError("Error getting factor values")


def get_data_files_command(
        study_id, output, json_query=None, galaxy_parameters_file=None):
    log.info("Getting data files for study %s. Writing to %s.",
             study_id, output.name)
    if json_query:
        log.debug("This is the specified query:\n%s", json_query)
        json_struct = json.loads(json_query)
        data_files = get_data_files(study_id, json_struct)
    elif galaxy_parameters_file:
        log.debug("Using input Galaxy JSON parameters from:\n%s",
                  galaxy_parameters_file)
        with open(galaxy_parameters_file) as json_fp:
            galaxy_json = json.load(json_fp)
            json_struct = {}
            for fv_item in galaxy_json['factor_value_series']:
                json_struct[fv_item['factor_name']] = fv_item['factor_value']
            data_files = get_data_files(study_id, json_struct)
    else:
        log.debug("No query was specified")
        data_files = get_data_files(study_id)

    log.debug("Result data files list: %s", data_files)
    if data_files is None:
        raise RuntimeError("Error getting data files with isatools")

    log.debug("dumping data files to %s", output.name)
    json.dump(list(data_files), output, indent=4)
    log.info("Finished writing data files to {}".format(output))


def build_html_data_files_list(data_files_list):
    data_files_table = '<table>'
    data_files_table += '<tr><th>Sample Name</th><th>Data File Names</th></tr>'
    for data_file in data_files_list:
        sample_name = data_file['sample']
        data_files = ', '.join(data_file['data_files'])
        data_files_table += '<tr><td>{sample_name}</td><td>{data_files}</td>' \
            .format(sample_name=sample_name, data_files=data_files)
    html_data_files_list = """
    <html>
    <head>
    <title>ISA-Tab Factors Summary</title>
    </head>
    <body>
    {summary_table}
    </body>
    </html>
""".format(summary_table=data_files_table)
    return html_data_files_list


def build_html_summary(summary):
    study_groups = {}
    for item in summary:
        sample_name = item['sample_name']
        study_factors = []
        for item in [x for x in item.items() if x[0] != "sample_name"]:
            study_factors.append(': '.join([item[0], item[1]]))
        study_group = ', '.join(study_factors)
        if study_group not in study_groups.keys():
            study_groups[study_group] = []
        study_groups[study_group].append(sample_name)
    summary_table = '<table>'
    summary_table += '<tr><th>Study group</th><th>Number of samples</th></tr>'
    for item in study_groups.items():
        study_group = item[0]
        num_samples = len(item[1])
        summary_table += '<tr><td>{study_group}</td><td>{num_samples}</td>' \
            .format(study_group=study_group, num_samples=num_samples)
    summary_table += '</table>'
    html_summary = """
<html>
<head>
<title>ISA-Tab Factors Summary</title>
</head>
<body>
{summary_table}
</body>
</html>
""".format(summary_table=summary_table)
    return html_summary


def get_summary_command(study_id, json_output, html_output):
    log.info("Getting summary for study %s. Writing to %s.",
             study_id, json_output.name)
    summary = get_study_variable_summary(study_id)
    if summary is not None:
        json.dump(summary, json_output, indent=4)
        log.debug("Summary dumped to JSON")
        html_summary = build_html_summary(summary)
        with html_output as html_fp:
            html_fp.write(html_summary)
    else:
        raise RuntimeError("Error getting study summary")


def datatype_get_summary_command(study_id, output):
    log.info("Getting summary for study %s. Writing to %s.",
             study_id, output.name)

    summary = get_study_variable_summary(study_id)
    print('summary: ', list(summary))
    if summary is not None:
        json.dump(summary, output, indent=4)
        log.debug("Summary dumped")
    else:
        raise RuntimeError("Error getting study summary")



"""Functions for reading, writing and validating ISA-JSON.

Don't forget to read the ISA-JSON spec:
http://isa-specs.readthedocs.io/en/latest/isajson.html
"""
from __future__ import absolute_import
import glob
import json
import logging
import os
import re
from io import StringIO
from json import JSONEncoder
from jsonschema import Draft4Validator, RefResolver, ValidationError

from isatools.model import (
    Investigation, OntologyAnnotation, Comment, OntologySource, Publication, Person, Study, Protocol, ProtocolParameter,
    ProtocolComponent, StudyFactor, Source, Characteristic, Sample, FactorValue, Process, ParameterValue, Assay,
    DataFile, Material,
)

__author__ = 'djcomlab@gmail.com (David Johnson)'


log = logging.getLogger('isatools')

errors = []
warnings = []
info = []

# REGEXES
_RX_DOI = re.compile("(10[.][0-9]{4,}(?:[.][0-9]+)*/(?:(?![%'#? ])\\S)+)")
_RX_PMID = re.compile("[0-9]{8}")
_RX_PMCID = re.compile("PMC[0-9]{8}")


def load(fp):

    def get_comments(commentable_dict):
        comments = [
            Comment(
                name=comm_dict.get('name', ''),
                value=comm_dict.get('value', '')
            ) for comm_dict in commentable_dict.get("comments", [])
        ]
        return comments

    def get_roles(j):
        roles = None
        if "roles" in j.keys():
            roles = list()
            for role_json in j["roles"]:
                term = role_json["annotationValue"]
                term_accession = role_json["termAccession"]
                term_source = term_source_dict[role_json["termSource"]]
                role = OntologyAnnotation(term, term_source, term_accession)
                roles.append(role)
        return roles

    def get_characteristic_category(characteristics_cats_dict):
        res = OntologyAnnotation(
            id_=characteristics_cats_dict["@id"],
            term=characteristics_cats_dict["characteristicType"]["annotationValue"],
            term_source=term_source_dict[characteristics_cats_dict["characteristicType"]["termSource"]],
            term_accession=characteristics_cats_dict["characteristicType"]["termAccession"],
        )
        try:
            res.comments = get_comments(characteristics_cats_dict)
        except KeyError:
            pass
        return res

    def get_parameter_value(p_val_dict):
        res = ParameterValue(
            category=parameters_dict[p_val_dict["category"]["@id"]],
            comments=get_comments(p_val_dict)
        )
        try:
            res.value = OntologyAnnotation(
                term=p_val_dict["value"]["annotationValue"],
                term_accession=p_val_dict["value"]["termAccession"],
                term_source=term_source_dict[p_val_dict["value"]["termSource"]],
                comments=get_comments(p_val_dict["value"])
            )
        except TypeError:
            res.value = p_val_dict["value"]
        return res

    def get_jvalue(dict, key):
        if key in dict.keys():
            return dict[key]
        else:
            return None

    investigation_json = json.load(fp)
    investigation = Investigation(
        identifier=investigation_json["identifier"],
        title=investigation_json["title"],
        description=investigation_json["description"],
        submission_date=investigation_json["submissionDate"],
        public_release_date=investigation_json["publicReleaseDate"]
    )
    investigation.comments = get_comments(investigation_json)
    term_source_dict = {"": None}
    for ontologySourceReference_json in investigation_json["ontologySourceReferences"]:
        ontology_source_reference = OntologySource(
            name=ontologySourceReference_json["name"],
            file=ontologySourceReference_json["file"],
            version=ontologySourceReference_json["version"],
            description=ontologySourceReference_json["description"],
            comments=get_comments(ontologySourceReference_json)
        )
        term_source_dict[ontology_source_reference.name] = ontology_source_reference
        investigation.ontology_source_references.append(ontology_source_reference)
    for publication_json in investigation_json["publications"]:
        publication = Publication(
            pubmed_id=publication_json["pubMedID"],
            doi=publication_json["doi"],
            author_list=publication_json["authorList"],
            title=publication_json["title"],
            status=OntologyAnnotation(
                term=publication_json["status"]["annotationValue"],
                term_accession=publication_json["status"]["termAccession"],
                term_source=term_source_dict[publication_json["status"]["termSource"]]
            )
        )
        try:
            publication.comments = get_comments(publication_json)
        except KeyError:
            pass
        investigation.publications.append(publication)
    for person_json in investigation_json["people"]:
        person = Person(
            last_name=person_json["lastName"],
            first_name=person_json["firstName"],
            mid_initials=person_json["midInitials"],
            email=person_json["email"],
            phone=person_json["phone"],
            fax=person_json["fax"],
            address=person_json["address"],
            affiliation=person_json["affiliation"],
            roles=[]
        )
        for role_json in person_json["roles"]:
            role = OntologyAnnotation(
                term=role_json["annotationValue"],
                term_accession=role_json["termAccession"],
                term_source=term_source_dict[role_json["termSource"]]
            )
            role.comments = get_comments(role_json)
            person.roles.append(role)
        person.comments = get_comments(person_json)
        investigation.contacts.append(person)
    samples_dict = dict()
    sources_dict = dict()
    categories_dict = dict()
    protocols_dict = dict()
    factors_dict = dict()
    parameters_dict = dict()
    units_dict = dict()

    # populate assay characteristicCategories first
    for study_json in investigation_json["studies"]:
        for assay_json in study_json["assays"]:
            for assay_characteristics_category_json in assay_json["characteristicCategories"]:
                characteristic_category = get_characteristic_category(assay_characteristics_category_json)
                # study.characteristic_categories.append(characteristic_category)
                categories_dict[characteristic_category.id] = characteristic_category
    for study_json in investigation_json["studies"]:
        process_dict = dict()
        study = Study(
            identifier=study_json["identifier"],
            title=study_json["title"],
            description=study_json["description"],
            submission_date=study_json["submissionDate"],
            public_release_date=study_json["publicReleaseDate"],
            filename=study_json["filename"]
        )
        try:
            study.comments = get_comments(study_json)
        except KeyError:
            pass
        for study_characteristics_category_json in study_json["characteristicCategories"]:
            characteristic_category = get_characteristic_category(study_characteristics_category_json)
            categories_dict[characteristic_category.id] = characteristic_category
            study.characteristic_categories.append(characteristic_category)
        for study_unit_json in study_json["unitCategories"]:
            unit = OntologyAnnotation(id_=study_unit_json["@id"],
                                      term=study_unit_json["annotationValue"],
                                      term_source=term_source_dict[study_unit_json["termSource"]],
                                      term_accession=study_unit_json["termAccession"])
            units_dict[unit.id] = unit
            study.units.append(unit)
        for study_publication_json in study_json["publications"]:
            study_publication = Publication(
                pubmed_id=study_publication_json["pubMedID"],
                doi=study_publication_json["doi"],
                author_list=study_publication_json["authorList"],
                title=study_publication_json["title"],
                status=OntologyAnnotation(
                    term=study_publication_json["status"]["annotationValue"],
                    term_source=term_source_dict[study_publication_json["status"]["termSource"]],
                    term_accession=study_publication_json["status"]["termAccession"],
                )
            )
            try:
                study_publication.comments = get_comments(study_publication_json)
            except KeyError:
                pass
            study.publications.append(study_publication)
        for study_person_json in study_json["people"]:
            study_person = Person(
                last_name=study_person_json["lastName"],
                first_name=study_person_json["firstName"],
                mid_initials=study_person_json["midInitials"],
                email=study_person_json["email"],
                phone=study_person_json["phone"],
                fax=study_person_json["fax"],
                address=study_person_json["address"],
                affiliation=study_person_json["affiliation"],
            )
            study_person.roles = get_roles(study_person_json)
            try:
                study_person.comments = get_comments(study_person_json)
            except KeyError:
                pass
            study.contacts.append(study_person)
        for design_descriptor_json in study_json["studyDesignDescriptors"]:
            design_descriptor = OntologyAnnotation(
                term=design_descriptor_json["annotationValue"],
                term_accession=design_descriptor_json["termAccession"],
                term_source=term_source_dict[design_descriptor_json["termSource"]]
            )
            study.design_descriptors.append(design_descriptor)
        for protocol_json in study_json["protocols"]:
            protocol = Protocol(
                id_=protocol_json["@id"],
                name=protocol_json["name"],
                uri=protocol_json["uri"],
                description=protocol_json["description"],
                version=protocol_json["version"],
                protocol_type=OntologyAnnotation(
                    term=protocol_json["protocolType"]["annotationValue"],
                    term_accession=protocol_json["protocolType"]["termAccession"] if "termAccession" in protocol_json["protocolType"].keys() else "",
                    term_source=term_source_dict[protocol_json["protocolType"]["termSource"]] if "termSource" in protocol_json["protocolType"].keys() else None,
                )
            )
            for parameter_json in protocol_json["parameters"]:
                parameter = ProtocolParameter(
                    id_=parameter_json["@id"],
                    parameter_name=OntologyAnnotation(
                        term=parameter_json["parameterName"]["annotationValue"],
                        term_source=term_source_dict[parameter_json["parameterName"]["termSource"]],
                        term_accession=parameter_json["parameterName"]["termAccession"]
                    ),
                    comments=get_comments(parameter_json)
                )
                protocol.parameters.append(parameter)
                parameters_dict[parameter.id] = parameter
            for component_json in protocol_json["components"]:
                component = ProtocolComponent(
                    name=component_json["componentName"],
                    component_type=OntologyAnnotation(
                        term=component_json["componentType"]["annotationValue"],
                        term_source=term_source_dict[component_json["componentType"]["termSource"]],
                        term_accession=component_json["componentType"]["termAccession"]
                    ),
                    comments=get_comments(component_json)
                )
                protocol.components.append(component)
            study.protocols.append(protocol)
            protocols_dict[protocol.id] = protocol
        for factor_json in study_json["factors"]:
            factor = StudyFactor(
                id_=factor_json["@id"],
                name=factor_json["factorName"],
                factor_type=OntologyAnnotation(
                    term=factor_json["factorType"]["annotationValue"],
                    term_accession=factor_json["factorType"]["termAccession"],
                    term_source=term_source_dict[factor_json["factorType"]["termSource"]]
                ),
                comments=get_comments(factor_json)
            )
            study.factors.append(factor)
            factors_dict[factor.id] = factor
        for source_json in study_json["materials"]["sources"]:
            source = Source(
                id_=source_json["@id"],
                name=source_json["name"][7:],
                comments=get_comments(source_json)
            )
            for characteristic_json in source_json["characteristics"]:
                value = characteristic_json["value"]
                unit = None
                characteristic = Characteristic(
                    category=categories_dict[characteristic_json["category"]["@id"]],
                    comments=get_comments(characteristic_json)
                )
                if isinstance(value, dict):
                    try:
                        term = characteristic_json["value"]["annotationValue"]
                        if isinstance(term, (int, float)):
                            term = str(term)
                        value = OntologyAnnotation(
                            term=term,
                            term_source=term_source_dict[characteristic_json["value"]["termSource"]],
                            term_accession=characteristic_json["value"]["termAccession"],
                            comments=get_comments(characteristic_json["value"])
                        )
                    except KeyError as ke:
                        raise IOError("Can't create value as annotation: " + str(ke) + " \n object: " + str(characteristic_json))
                elif isinstance(value, (int, float)):
                    try:
                        unit = units_dict[characteristic_json["unit"]["@id"]]
                    except KeyError:
                        unit = None
                elif not isinstance(value, str):
                    raise IOError("Unexpected type in characteristic value")
                characteristic.value = value
                characteristic.unit = unit
                source.characteristics.append(characteristic)
            sources_dict[source.id] = source
            study.sources.append(source)
        for sample_json in study_json["materials"]["samples"]:
            sample = Sample(
                id_=sample_json["@id"],
                name=sample_json["name"][7:],
                comments=get_comments(sample_json)
            )
            for characteristic_json in sample_json["characteristics"]:
                value = characteristic_json["value"]
                unit = None
                characteristic = Characteristic(
                    category=categories_dict[characteristic_json["category"]["@id"]],
                    comments=get_comments(characteristic_json)
                )
                if isinstance(value, dict):
                    try:
                        value = OntologyAnnotation(
                            term=characteristic_json["value"]["annotationValue"],
                            term_source=term_source_dict[characteristic_json["value"]["termSource"]],
                            term_accession=characteristic_json["value"]["termAccession"],
                            comments=get_comments(characteristic_json["value"])
                        )
                    except KeyError as ke:
                        raise IOError("Can't create value as annotation: " + str(ke) + "\n object: " + str(characteristic_json))
                elif isinstance(value, int) or isinstance(value, float):
                    try:
                        unit = units_dict[characteristic_json["unit"]["@id"]]
                    except KeyError:
                        unit = None
                        # unit = characteristic_json.get('unit', None)
                elif not isinstance(value, str):
                    raise IOError("Unexpected type in characteristic value")
                characteristic.value = value
                characteristic.unit = unit
                sample.characteristics.append(characteristic)
            for factor_value_json in sample_json["factorValues"]:
                value = factor_value_json["value"]
                unit = None
                factor_value = FactorValue(
                    factor_name=factors_dict[factor_value_json["category"]["@id"]],
                    comments=get_comments(factor_value_json)
                )
                if isinstance(value, dict):
                    try:
                        value = OntologyAnnotation(
                            term=factor_value_json["value"]["annotationValue"],
                            term_accession=factor_value_json["value"]["termAccession"],
                            term_source=term_source_dict[factor_value_json["value"]["termSource"]],
                            comments=get_comments(factor_value_json["value"])
                        )
                    except KeyError as ke:
                        raise IOError("Can't create value as annotation: " + str(ke) + "\n object: " + str(factor_value_json))
                elif isinstance(value, (int, float)):
                    try:
                        unit = units_dict[factor_value_json["unit"]["@id"]]
                    except KeyError:
                        unit = None
                        # unit = factor_value_json.get("unit", None)
                elif not isinstance(value, str):
                    raise IOError("Unexpected type in factor value")
                factor_value.value = value
                factor_value.unit = unit
                sample.factor_values.append(factor_value)
            samples_dict[sample.id] = sample
            study.samples.append(sample)
            try:
                for source_id_ref_json in sample_json["derivesFrom"]:
                    sample.derives_from.append(sources_dict[source_id_ref_json["@id"]])
            except KeyError:
                sample.derives_from = []
        for study_process_json in study_json["processSequence"]:
            process = Process(
                id_=study_process_json["@id"],
                executes_protocol=protocols_dict[study_process_json["executesProtocol"]["@id"]],
            )
            try:
                process.comments = get_comments(study_process_json)
            except KeyError:
                pass
            try:
                process.date = study_process_json["date"]
            except KeyError:
                pass
            try:
                process.performer = study_process_json["performer"]
            except KeyError:
                pass
            for parameter_value_json in study_process_json["parameterValues"]:
                if isinstance(parameter_value_json["value"], int) or isinstance(parameter_value_json["value"], float):
                    parameter_value = ParameterValue(
                        category=parameters_dict[parameter_value_json["category"]["@id"]],
                        value=parameter_value_json["value"],
                        unit=units_dict[parameter_value_json["unit"]["@id"]],
                        comments=get_comments(parameter_value_json)
                    )
                    process.parameter_values.append(parameter_value)
                else:
                    parameter_value = get_parameter_value(parameter_value_json)
                    process.parameter_values.append(parameter_value)
            for input_json in study_process_json["inputs"]:
                input_ = None
                try:
                    input_ = sources_dict[input_json["@id"]]
                except KeyError:
                    pass
                finally:
                    try:
                        input_ = samples_dict[input_json["@id"]]
                    except KeyError:
                        pass
                if input_ is None:
                    raise IOError("Could not find input node in sources or samples dicts: " + input_json["@id"])
                process.inputs.append(input_)
            for output_json in study_process_json["outputs"]:
                output = None
                try:
                    output = sources_dict[output_json["@id"]]
                except KeyError:
                    pass
                finally:
                    try:
                        output = samples_dict[output_json["@id"]]
                    except KeyError:
                        pass
                if output is None:
                    raise IOError("Could not find output node in sources or samples dicts: " + output_json["@id"])
                process.outputs.append(output)
            study.process_sequence.append(process)
            process_dict[process.id] = process
        for study_process_json in study_json["processSequence"]:  # 2nd pass
            try:
                prev_proc = study_process_json["previousProcess"]["@id"]
                process_dict[study_process_json["@id"]].prev_process = process_dict[prev_proc]
            except KeyError:
                pass

            try:
                next_proc = study_process_json["nextProcess"]["@id"]
                process_dict[study_process_json["@id"]].next_process = process_dict[next_proc]
            except KeyError:
                pass

        for assay_json in study_json["assays"]:
            process_dict = dict()
            assay = Assay(
                measurement_type=OntologyAnnotation(
                    term=assay_json["measurementType"]["annotationValue"],
                    term_accession=assay_json["measurementType"]["termAccession"],
                    term_source=term_source_dict[assay_json["measurementType"]["termSource"]]
                ),
                technology_type=OntologyAnnotation(
                    term=assay_json["technologyType"]["annotationValue"],
                    term_accession=assay_json["technologyType"]["termAccession"],
                    term_source=term_source_dict[assay_json["technologyType"]["termSource"]]
                ),
                technology_platform=assay_json["technologyPlatform"],
                filename=assay_json["filename"],
                comments=get_comments(assay_json)
            )
            for assay_unit_json in assay_json["unitCategories"]:
                unit = OntologyAnnotation(
                    id_=assay_unit_json["@id"],
                    term=assay_unit_json["annotationValue"],
                    term_source=term_source_dict[assay_unit_json["termSource"]],
                    term_accession=assay_unit_json["termAccession"],
                    comments=get_comments(assay_unit_json)
                )
                units_dict[unit.id] = unit
                assay.units.append(unit)
            data_dict = dict()
            for data_json in assay_json["dataFiles"]:
                data_file = DataFile(
                    id_=data_json["@id"],
                    filename=data_json["name"],
                    label=data_json["type"],
                    comments=get_comments(data_json)
                )
                data_dict[data_file.id] = data_file
                try:
                    data_file.derives_from = samples_dict[data_json["derivesFrom"][0]["@id"]]
                except KeyError:
                    data_file.derives_from = None
                assay.data_files.append(data_file)
            for sample_json in assay_json["materials"]["samples"]:
                sample = samples_dict[sample_json["@id"]]
                assay.samples.append(sample)
            for assay_characteristics_category_json in assay_json["characteristicCategories"]:
                characteristic_category = OntologyAnnotation(
                    id_=assay_characteristics_category_json["@id"],
                    term=assay_characteristics_category_json["characteristicType"]["annotationValue"],
                    term_source=term_source_dict[assay_characteristics_category_json["characteristicType"]["termSource"]],
                    term_accession=assay_characteristics_category_json["characteristicType"]["termAccession"],
                    comments=get_comments(assay_characteristics_category_json["characteristicType"])
                )
                study.characteristic_categories.append(characteristic_category)
                categories_dict[characteristic_category.id] = characteristic_category
            other_materials_dict = dict()
            for other_material_json in assay_json["materials"]["otherMaterials"]:
                material_name = other_material_json["name"]
                if material_name.startswith("labeledextract-"):
                    material_name = material_name[15:]
                else:
                    material_name = material_name[8:]
                material = Material(
                    id_=other_material_json["@id"],
                    name=material_name,
                    type_=other_material_json["type"],
                    comments=get_comments(other_material_json)
                )
                for characteristic_json in other_material_json["characteristics"]:
                    characteristic = Characteristic(
                        category=categories_dict[characteristic_json["category"]["@id"]],
                        value=OntologyAnnotation(
                            term=characteristic_json["value"]["annotationValue"],
                            term_source=term_source_dict[characteristic_json["value"]["termSource"]],
                            term_accession=characteristic_json["value"]["termAccession"],
                            comments=get_comments(characteristic_json["value"])
                        ),
                        comments=get_comments(characteristic_json)
                    )
                    material.characteristics.append(characteristic)
                assay.other_material.append(material)
                other_materials_dict[material.id] = material
            for assay_process_json in assay_json["processSequence"]:
                process = Process(
                    id_=assay_process_json["@id"],
                    executes_protocol=protocols_dict[assay_process_json["executesProtocol"]["@id"]],
                    comments=get_comments(assay_process_json)
                )
                # additional properties, currently hard-coded special cases
                if process.executes_protocol.protocol_type.term == "data collection" and assay.technology_type.term == "DNA microarray":
                    process.name = assay_process_json["name"]
                elif process.executes_protocol.protocol_type.term == "nucleic acid sequencing":
                    process.name = assay_process_json["name"]
                elif process.executes_protocol.protocol_type.term == "nucleic acid hybridization":
                    process.name = assay_process_json["name"]
                elif process.executes_protocol.protocol_type.term == "data transformation":
                    process.name = assay_process_json["name"]
                elif process.executes_protocol.protocol_type.term == "data normalization":
                    process.name = assay_process_json["name"]
                for input_json in assay_process_json["inputs"]:
                    input_ = None
                    try:
                        input_ = samples_dict[input_json["@id"]]
                    except KeyError:
                        pass
                    finally:
                        try:
                            input_ = other_materials_dict[input_json["@id"]]
                        except KeyError:
                            pass
                        finally:
                            try:
                                input_ = data_dict[input_json["@id"]]
                            except KeyError:
                                pass
                    if input_ is None:
                        raise IOError("Could not find input node in samples or materials or data dicts: " +
                                      input_json["@id"])
                    process.inputs.append(input_)
                for output_json in assay_process_json["outputs"]:
                    output = None
                    try:
                        output = samples_dict[output_json["@id"]]
                    except KeyError:
                        pass
                    finally:
                        try:
                            output = other_materials_dict[output_json["@id"]]
                        except KeyError:
                            pass
                        finally:
                                try:
                                    output = data_dict[output_json["@id"]]
                                except KeyError:
                                    pass
                    if output is None:
                        raise IOError("Could not find output node in samples or materials or data dicts: " +
                                      output_json["@id"])
                    process.outputs.append(output)
                for parameter_value_json in assay_process_json["parameterValues"]:
                    if "category" in parameter_value_json.keys():
                        if parameter_value_json["category"]["@id"] == "#parameter/Array_Design_REF":  # Special case
                            process.array_design_ref = parameter_value_json["value"]
                        elif isinstance(parameter_value_json["value"], int) or \
                                isinstance(parameter_value_json["value"], float):
                            parameter_value = ParameterValue(
                                category=parameters_dict[parameter_value_json["category"]["@id"]],
                                value=parameter_value_json["value"],
                                comments=get_comments(parameter_value_json)
                            )
                            if "unit" in parameter_value_json.keys():
                                parameter_value.unit = units_dict[parameter_value_json["unit"]["@id"]]
                            process.parameter_values.append(parameter_value)
                        else:
                            parameter_value = get_parameter_value(parameter_value_json)
                            process.parameter_values.append(parameter_value)
                    else:
                        log.warning("warning: parameter category not found for instance {}".format(parameter_json))
                assay.process_sequence.append(process)
                process_dict[process.id] = process

                for assay_process_json in assay_json["processSequence"]:  # 2nd pass
                    try:
                        prev_proc = assay_process_json["previousProcess"]["@id"]
                        process_dict[assay_process_json["@id"]].prev_process = process_dict[prev_proc]
                    except KeyError:
                        pass

                    try:
                        next_proc = assay_process_json["nextProcess"]["@id"]
                        process_dict[assay_process_json["@id"]].next_process = process_dict[next_proc]
                    except KeyError:
                        pass

            study.assays.append(assay)
        investigation.studies.append(study)
    return investigation


"""Everything below here is for the validator"""


def get_source_ids(study_json):
    """Used for rule 1002"""
    return [source["@id"] for source in study_json["materials"]["sources"]]


def get_sample_ids(study_json):
    """Used for rule 1003"""
    return [sample["@id"] for sample in study_json["materials"]["samples"]]


def get_material_ids(study_json):
    """Used for rule 1005"""
    material_ids = list()
    for assay_json in study_json["assays"]:
        material_ids.extend([material["@id"] for material in assay_json["materials"]["otherMaterials"]])
    return material_ids


def get_data_file_ids(study_json):
    """Used for rule 1004"""
    data_file_ids = list()
    for assay_json in study_json["assays"]:
        data_file_ids.extend([data_file["@id"] for data_file in assay_json["dataFiles"]])
    return data_file_ids


def get_io_ids_in_process_sequence(study_json):
    """Used for rules 1001-1005"""
    all_process_sequences = list(study_json["processSequence"])
    for assay_json in study_json["assays"]:
        all_process_sequences.extend(assay_json["processSequence"])
    return [elem for iterabl in [[i["@id"] for i in process["inputs"]] + [o["@id"] for o in process["outputs"]] for process in
                                 all_process_sequences] for elem in iterabl]


def check_material_ids_declared_used(study_json, id_collector_func):
    """Used for rules 1015-1018"""
    node_ids = id_collector_func(study_json)
    io_ids_in_process_sequence = get_io_ids_in_process_sequence(study_json)
    is_node_ids_used = set(node_ids).issubset(set(io_ids_in_process_sequence))
    if not is_node_ids_used:
        warnings.append({
            "message": "Material declared but not used",
            "supplemental": "{} not used in any inputs/outputs in {}".format(node_ids, io_ids_in_process_sequence),
            "code": 1017
        })
        log.warning("(W) Not all node IDs in {} used by inputs/outputs {}".format(node_ids,
                                                                                  io_ids_in_process_sequence))


def check_material_ids_not_declared_used(study_json):
    """Used for rules 1002-1005"""
    node_ids = get_source_ids(study_json) + get_sample_ids(study_json) + get_material_ids(study_json) + \
               get_data_file_ids(study_json)
    io_ids_in_process_sequence = get_io_ids_in_process_sequence(study_json)
    if len(set(io_ids_in_process_sequence)) - len(set(node_ids)) > 0:
        diff = set(io_ids_in_process_sequence) - set(node_ids)
        errors.append({
            "message": "Missing Material",
            "supplemental": "Inputs/outputs in {}  not found in sources, samples, materials or datafiles "
                            "declarations".format(list(diff)),
            "code": 1005
        })
        log.error("(E) There are some inputs/outputs IDs {} not found in sources, samples, materials or data files"
                     "declared".format(list(diff)))


def check_process_sequence_links(process_sequence_json):
    """Used for rule 1006"""
    process_ids = [process["@id"] for process in process_sequence_json]
    for process in process_sequence_json:
        try:
            if process["previousProcess"]["@id"] not in process_ids:
                errors.append({
                    "message": "Missing Process link",
                    "supplemental": "previousProcess {} in process {} does not refer to another process in "
                                    "sequence".format(process["previousProcess"]["@id"], process["@id"]),
                    "code": 1006
                })
                log.error("(E) previousProcess link {} in process {} does not refer to another process in "
                             "sequence".format(process["previousProcess"]["@id"], process["@id"]))
        except KeyError:
            pass
        try:
            if process["nextProcess"]["@id"] not in process_ids:
                errors.append({
                    "message": "Missing Process link",
                    "supplemental": "nextProcess {} in process {} does not refer to another process in "
                                    "sequence".format(process["nextProcess"]["@id"], process["@id"]),
                    "code": 1006
                })
                log.error("(E) nextProcess {} in process {} does not refer to another process in sequence".format(
                    process["nextProcess"]["@id"], process["@id"]))
        except KeyError:
            pass


def get_study_protocol_ids(study_json):
    """Used for rule 1007"""
    return [protocol["@id"] for protocol in study_json["protocols"]]


def check_process_protocol_ids_usage(study_json):
    """Used for rules 1007 and 1019"""
    protocol_ids_declared = get_study_protocol_ids(study_json)
    process_sequence = study_json["processSequence"]
    protocol_ids_used = list()
    for process in process_sequence:
        try:
            protocol_ids_used.append(process["executesProtocol"]["@id"])
        except KeyError:
            pass
    for assay in study_json["assays"]:
        process_sequence = assay["processSequence"]
        for process in process_sequence:
            try:
                protocol_ids_used.append(process["executesProtocol"]["@id"])
            except KeyError:
                pass
    if len(set(protocol_ids_used) - set(protocol_ids_declared)) > 0:
        diff = set(protocol_ids_used) - set(protocol_ids_declared)
        errors.append({
            "message": "Missing Protocol declaration",
            "supplemental": "protocol IDs {} not declared".format(list(diff)),
            "code": 1007
        })
        log.error("(E) There are protocol IDs {} used in a study or assay process sequence not declared".format(
            list(diff)))
    elif len(set(protocol_ids_declared) - set(protocol_ids_used)) > 0:
        diff = set(protocol_ids_declared) - set(protocol_ids_used)
        warnings.append({
            "message": "Protocol declared but not used",
            "supplemental": "protocol IDs declared {} not used".format(list(diff)),
            "code": 1019
        })
        log.warning("(W) There are some protocol IDs declared {} not used in any study or assay process "
                       "sequence".format(list(diff)))


def get_study_protocols_parameter_ids(study_json):
    """Used for rule 1009"""
    return [elem for iterabl in [[param["@id"] for param in protocol["parameters"]] for protocol in
                                 study_json["protocols"]] for elem in iterabl]


def get_parameter_value_parameter_ids(study_json):
    """Used for rule 1009"""
    study_pv_parameter_ids = [elem for iterabl in
                              [[parameter_value["category"]["@id"] for parameter_value in process["parameterValues"]]
                               for process in study_json["processSequence"]] for elem in iterabl]
    for assay in study_json["assays"]:
        study_pv_parameter_ids.extend([elem for iterabl in
                                       [[parameter_value["category"]["@id"] for parameter_value in
                                         process["parameterValues"]]
                                        for process in assay["processSequence"]] for elem in iterabl]
                                      )
    return study_pv_parameter_ids


def check_protocol_parameter_ids_usage(study_json):
    """Used for rule 1009 and 1020"""
    protocols_declared = get_study_protocols_parameter_ids(study_json) + ["#parameter/Array_Design_REF"] # + special case
    protocols_used = get_parameter_value_parameter_ids(study_json)
    if len(set(protocols_used) - set(protocols_declared)) > 0:
        diff = set(protocols_used) - set(protocols_declared)
        errors.append({
            "message": "Missing Protocol Parameter declaration",
            "supplemental": "protocol parameters {} used".format(list(diff)),
            "code": 1009
        })
        log.error("(E) There are protocol parameters {} used in a study or assay process not declared in any "
                     "protocol".format(list(diff)))
    elif len(set(protocols_declared) - set(protocols_used)) > 0:
        diff = set(protocols_declared) - set(protocols_used)
        warnings.append({
            "message": "Protocol parameter declared in a protocol but never used",
            "supplemental": "protocol declared {} are not used".format(list(diff)),
            "code": 1020
        })
        log.warning("(W) There are some protocol parameters declared {} not used in any study or assay process"
                    .format(list(diff)))


def get_characteristic_category_ids(study_or_assay_json):
    """Used for rule 1013"""
    return [category["@id"] for category in study_or_assay_json["characteristicCategories"]]


def get_characteristic_category_ids_in_study_materials(study_json):
    """Used for rule 1013"""
    return [elem for iterabl in
            [[characteristic["category"]["@id"] for characteristic in material["characteristics"]] for material in
             study_json["materials"]["sources"] + study_json["materials"]["samples"]] for elem in iterabl]


def get_characteristic_category_ids_in_assay_materials(assay_json):
    """Used for rule 1013"""
    return [elem for iterabl in [[characteristic["category"]["@id"]  for characteristic in material["characteristics"]]
                                 if "characteristics" in material.keys() else [] for material in
              assay_json["materials"]["samples"] + assay_json["materials"]["otherMaterials"]] for elem in iterabl]


def check_characteristic_category_ids_usage(studies_json):
    """Used for rule 1013"""
    characteristic_categories_declared = list()
    characteristic_categories_used = list()
    for study_json in studies_json:
        characteristic_categories_declared += get_characteristic_category_ids(study_json)
        for assay in study_json["assays"]:
            characteristic_categories_declared_in_assay = get_characteristic_category_ids(assay)
            characteristic_categories_declared += characteristic_categories_declared_in_assay
        characteristic_categories_used += get_characteristic_category_ids_in_study_materials(study_json)
        for assay in study_json["assays"]:
            characteristic_categories_used_in_assay = get_characteristic_category_ids_in_assay_materials(assay)
            characteristic_categories_used += characteristic_categories_used_in_assay
    if len(set(characteristic_categories_used) - set(characteristic_categories_declared)) > 0:
        diff = set(characteristic_categories_used) - set(characteristic_categories_declared)
        errors.append({
                "message": "Missing Characteristic Category declaration",
                "supplemental": "Characteristic Categories {} used not declared".format(list(diff)),
                "code": 1013
            })
        log.error("(E) There are characteristic categories {} used in a source or sample characteristic that have "
                     "not been not declared".format(list(diff)))
    elif len(set(characteristic_categories_declared) - set(characteristic_categories_used)) > 0:
        diff = set(characteristic_categories_declared) - set(characteristic_categories_used)
        warnings.append({
            "message": "Characteristic Category not used",
            "supplemental": "Characteristic Categories {} declared".format(list(diff)),
            "code": 1022
        })
        log.warning("(W) There are characteristic categories declared {} that have not been used in any source or "
                       "sample characteristic".format(list(diff)))


def get_study_factor_ids(study_json):
    """Used for rule 1008 and 1021"""
    return [factor["@id"] for factor in study_json["factors"]]


def get_study_factor_ids_in_sample_factor_values(study_json):
    """Used for rule 1008 and 1021"""
    return [elem for iterabl in [[factor["category"]["@id"] for factor in sample["factorValues"]] for sample in
                                 study_json["materials"]["samples"]] for elem in iterabl]


def check_study_factor_usage(study_json):
    """Used for rules 1008 and 1021"""
    factors_declared = get_study_factor_ids(study_json)
    factors_used = get_study_factor_ids_in_sample_factor_values(study_json)
    if len(set(factors_used) - set(factors_declared)) > 0:
        diff = set(factors_used) - set(factors_declared)
        errors.append({
            "message": "Missing Study Factor declaration",
            "supplemental": "Study Factors {} used".format(list(diff)),
            "code": 1008
        })
        log.error("(E) There are study factors {} used in a sample factor value that have not been not declared"
                  .format(list(diff)))
    elif len(set(factors_declared) - set(factors_used)) > 0:
        diff = set(factors_declared) - set(factors_used)
        warnings.append({
            "message": "Study Factor is not used",
            "supplemental": "Study Factors {} are not used".format(list(diff)),
            "code": 1021
        })
        log.warning("(W) There are some study factors declared {} that have not been used in any sample factor value"
                    .format(list(diff)))


def get_unit_category_ids(study_or_assay_json):
    """Used for rule 1014"""
    return [category["@id"] for category in study_or_assay_json["unitCategories"]]


def get_study_unit_category_ids_in_materials_and_processes(study_json):
    """Used for rule 1014"""
    study_characteristics_units_used = [elem for iterabl in
                                        [[characteristic["unit"]["@id"] if "unit" in characteristic.keys() else None for
                                          characteristic in material["characteristics"]] for material in
                                         study_json["materials"]["sources"] + study_json["materials"]["samples"]] for
                                        elem in iterabl]
    study_factor_value_units_used = [elem for iterabl in
                                     [[factor_value["unit"]["@id"] if "unit" in factor_value.keys() else None for
                                       factor_value in material["factorValues"]] for material in
                                      study_json["materials"]["samples"]] for
                                     elem in iterabl]
    parameter_value_units_used = [elem for iterabl in[[parameter_value["unit"]["@id"]
                                                       if "unit" in parameter_value.keys() else None for
                                   parameter_value in process["parameterValues"]] for process in
                                  study_json["processSequence"]] for
                                  elem in iterabl]
    return [x for x in study_characteristics_units_used + study_factor_value_units_used + parameter_value_units_used
            if x is not None]


def get_assay_unit_category_ids_in_materials_and_processes(assay_json):
    """Used for rule 1014"""
    assay_characteristics_units_used = [elem for iterabl in [[characteristic["unit"]["@id"] if "unit" in
                                        characteristic.keys() else None
                                                              for characteristic in material["characteristics"]]
                                                             if "characteristics" in material.keys() else None for
                                     material in assay_json["materials"]["otherMaterials"]] for elem in iterabl]
    parameter_value_units_used = [elem for iterabl in[[parameter_value["unit"]["@id"]
                                                       if "unit" in parameter_value.keys() else None
                                                       for parameter_value in process["parameterValues"]] for process in
                                                      assay_json["processSequence"]] for
                                  elem in iterabl]
    return [x for x in assay_characteristics_units_used + parameter_value_units_used if x is not None]


def check_unit_category_ids_usage(study_json):
    """Used for rules 1014 and 1022"""
    log.info("Getting units declared...")
    units_declared = get_unit_category_ids(study_json)
    for assay in study_json["assays"]:
        units_declared.extend(get_unit_category_ids(assay))
    log.info("Getting units used (study)...")
    units_used = get_study_unit_category_ids_in_materials_and_processes(study_json)
    log.info("Getting units used (assay)...")
    for assay in study_json["assays"]:
        units_used.extend(get_assay_unit_category_ids_in_materials_and_processes(assay))
    log.info("Comparing units declared vs units used...")
    if len(set(units_used) - set(units_declared)) > 0:
        diff = set(units_used) - set(units_declared)
        log.error("(E) There are units {} used in a material or parameter value that have not been not declared"
                  .format(list(diff)))
    elif len(set(units_declared) - set(units_used)) > 0:
        diff = set(units_declared) - set(units_used)
        warnings.append({
            "message": "Unit declared but not used",
            "supplemental": "Units declared {} not used".format(list(diff)),
            "code": 1022
        })
        log.warning("(W) There are some units declared {} that have not been used in any material or parameter value"
                    .format(list(diff)))


def check_utf8(fp):
    """Used for rule 0010"""
    import chardet
    with open(fp.name, "rb") as fp:
        charset = chardet.detect(fp.read())
        if charset["encoding"] != "UTF-8" and charset["encoding"] != "ascii":
            warnings.append({
                "message": "File should be UTF8 encoding",
                "supplemental": "Encoding is '{0}' with confidence {1}".format(charset["encoding"], charset["confidence"]),
                "code": 10
            })
            log.warning("(W) File should be UTF-8 encoding but found it is '{0}' encoding with {1} confidence"
                        .format(charset["encoding"], charset["confidence"]))
            raise SystemError()


def check_isa_schemas(isa_json, investigation_schema_path):
    """Used for rule 0003 and 4003"""
    try:
        with open(investigation_schema_path) as fp:
            investigation_schema = json.load(fp)
            resolver = RefResolver("file://" + investigation_schema_path, investigation_schema)
            validator = Draft4Validator(investigation_schema, resolver=resolver)
            validator.validate(isa_json)
    except ValidationError as ve:
        errors.append({
            "message": "Invalid JSON against ISA-JSON schemas",
            "supplemental": str(ve),
            "code": 3
        })
        log.fatal("(F) The JSON does not validate against the provided ISA-JSON schemas!")
        log.fatal("Fatal error: " + str(ve))
        raise SystemError("(F) The JSON does not validate against the provided ISA-JSON schemas!")


def check_date_formats(isa_json):
    """Used for rule 3001"""
    def check_iso8601_date(date_str):
        if date_str != "":
            try:
                iso8601.parse_date(date_str)
            except iso8601.ParseError:
                warnings.append({
                    "message": "Date is not ISO8601 formatted",
                    "supplemental": "Found {} in date field".format(date_str),
                    "code": 3001
                })
                log.warning("(W) Date {} does not conform to ISO8601 format".format(date_str))
    import iso8601
    try:
        check_iso8601_date(isa_json["publicReleaseDate"])
    except KeyError:
        pass
    try:
        check_iso8601_date(isa_json["submissionDate"])
    except KeyError:
        pass
    for study in isa_json["studies"]:
        try:
            check_iso8601_date(study["publicReleaseDate"])
        except KeyError:
            pass
        try:
            check_iso8601_date(study["submissionDate"])
        except KeyError:
            pass
        for process in study["processSequence"]:
            try:
                check_iso8601_date(process["date"])
            except KeyError:
                pass


def check_dois(isa_json):
    """Used for rule 3002"""
    def check_doi(doi_str):
        if doi_str != "":
            if not _RX_DOI.match(doi_str):
                warnings.append({
                    "message": "DOI is not valid format",
                    "supplemental": "Found {} in DOI field".format(doi_str),
                    "code": 3002
                })
                log.warning("(W) DOI {} does not conform to DOI format".format(doi_str))
    for ipub in isa_json["publications"]:
        try:
            check_doi(ipub["doi"])
        except KeyError:
            pass
    for study in isa_json["studies"]:
        for spub in study["publications"]:
            try:
                check_doi(spub["doi"])
            except KeyError:
                pass


def check_filenames_present(isa_json):
    """Used for rule 3005"""
    for s_pos, study in enumerate(isa_json["studies"]):
        if study["filename"] == "":
            warnings.append({
                "message": "Missing study file name",
                "supplemental": "At study position {}".format(s_pos),
                "code": 3005
            })
            log.warning("(W) A study filename is missing")
        for a_pos, assay in enumerate(study["assays"]):
            if assay["filename"] == "":
                warnings.append({
                    "message": "Missing assay file name",
                    "supplemental": "At study position {}, assay position {}".format(s_pos, a_pos),
                    "code": 3005
                })
                log.warning("(W) An assay filename is missing")


def check_pubmed_ids_format(isa_json):
    """Used for rule 3003"""
    def check_pubmed_id(pubmed_id_str):
        if pubmed_id_str != "":
            if (_RX_PMID.match(pubmed_id_str) is None) and (_RX_PMCID.match(pubmed_id_str) is None):
                warnings.append({
                    "message": "PubMed ID is not valid format",
                    "supplemental": "Found PubMedID {}".format(pubmed_id_str),
                    "code": 3003
                })
                log.warning("(W) PubMed ID {} is not valid format".format(pubmed_id_str))
    for ipub in isa_json["publications"]:
        check_pubmed_id(ipub["pubMedID"])
    for study in isa_json["studies"]:
        for spub in study["publications"]:
            check_pubmed_id(spub["pubMedID"])


def check_protocol_names(isa_json):
    """Used for rule 1010"""
    for study in isa_json["studies"]:
        for protocol in study["protocols"]:
            if protocol["name"] == "":
                warnings.append({
                    "message": "Protocol missing name",
                    "supplemental": "Protocol @id={}".format(protocol["@id"]),
                    "code": 1010
                })
                log.warning("(W) A Protocol {} is missing Protocol Name, so can't be referenced in ISA-tab"
                            .format(protocol["@id"]))


def check_protocol_parameter_names(isa_json):
    """Used for rule 1011"""
    for study in isa_json["studies"]:
        for protocol in study["protocols"]:
            for parameter in protocol["parameters"]:
                if parameter["parameterName"] == "":
                    warnings.append({
                        "message": "Protocol Parameter missing name",
                        "supplemental": "Protocol Parameter @id={}".format(parameter["@id"]),
                        "code": 1011
                    })
                    log.warning("(W) A Protocol Parameter {} is missing name, so can't be referenced in ISA-tab"
                                .format(parameter["@id"]))


def check_study_factor_names(isa_json):
    """Used for rule 1012"""
    for study in isa_json["studies"]:
        for factor in study["factors"]:
            if factor["factorName"] == "":
                warnings.append({
                    "message": "Study Factor missing name",
                    "supplemental": "Study Factor @id={}".format(factor["@id"]),
                    "code": 1012
                })
                log.warning("(W) A Study Factor is missing name, so can't be referenced in ISA-tab"
                            .format(factor["@id"]))


def check_ontology_sources(isa_json):
    """Used for rule 3008"""
    for ontology_source in isa_json["ontologySourceReferences"]:
        if ontology_source["name"] == "":
            warnings.append({
                "message": "Ontology Source missing name ref",
                "supplemental": "name={}".format(ontology_source["name"]),
                "code": 3008
            })
            log.warning("(W) An Ontology Source Reference is missing Term Source Name, so can't be referenced")


def get_ontology_source_refs(isa_json):
    """Used for rules 3007 and 3009"""
    return [ontology_source_ref["name"] for ontology_source_ref in isa_json["ontologySourceReferences"]]


def walk_and_get_annotations(isa_json, collector):
    """Used for rules 3007 and 3009

    Usage:
      collector = list()
      walk_and_get_annotations(isa_json, collector)
      # and then like magic all your annotations from the JSON should be in the collector list
    """
    #  Walk JSON tree looking for ontology annotation structures in the JSON
    if isinstance(isa_json, dict):
        if set(isa_json.keys()) == {"annotationValue", "termAccession", "termSource"} or \
                        set(isa_json.keys()) == {"@id", "annotationValue", "termAccession", "termSource"}:
            collector.append(isa_json)
        for i in isa_json.keys():
            walk_and_get_annotations(isa_json[i], collector)
    elif isinstance(isa_json, list):
        for j in isa_json:
            walk_and_get_annotations(j, collector)


def check_term_source_refs(isa_json):
    """Used for rules 3007 and 3009"""
    term_sources_declared = get_ontology_source_refs(isa_json)
    collector = list()
    walk_and_get_annotations(isa_json, collector)
    term_sources_used = [annotation["termSource"] for annotation in collector if annotation["termSource"] != ""]
    if len(set(term_sources_used) - set(term_sources_declared)) > 0:
        diff = set(term_sources_used) - set(term_sources_declared)
        errors.append({
            "message": "Missing Term Source",
            "supplemental": "Ontology sources missing {}".format(list(diff)),
            "code": 3009
        })
        log.error("(E) There are ontology sources {} referenced in an annotation that have not been not declared"
                  .format(list(diff)))
    elif len(set(term_sources_declared) - set(term_sources_used)) > 0:
        diff = set(term_sources_declared) - set(term_sources_used)
        warnings.append({
            "message": "Ontology Source Reference != used",
            "supplemental": "Ontology sources not used {}".format(list(diff)),
            "code": 3007
        })
        log.warning("(W) There are some ontology sources declared {} that have not been used in any annotation"
                    .format(list(diff)))


def check_term_accession_used_no_source_ref(isa_json):
    """Used for rule 3010"""
    collector = list()
    walk_and_get_annotations(isa_json, collector)
    terms_using_accession_no_source_ref = [
        annotation for annotation in collector if annotation["termAccession"] != "" and annotation["termSource"] == ""
    ]
    if len(terms_using_accession_no_source_ref) > 0:
        warnings.append({
            "message": "Missing Term Source REF in annotation",
            "supplemental": "Terms with accession but no source reference {}".format(terms_using_accession_no_source_ref),
            "code": 3010
        })
        log.warning("(W) There are ontology annotations with termAccession set but no termSource referenced: {}"
                    .format(terms_using_accession_no_source_ref))


def load_config(config_dir):
    import json
    configs = dict()
    for file in glob.iglob(os.path.join(config_dir, "*.json")):
        try:
            with open(file) as fp:
                config_dict = json.load(fp)
                if os.path.basename(file) == "protocol_definitions.json":
                    configs["protocol_definitions"] = config_dict
                elif os.path.basename(file) == "study_config.json":
                    configs["study"] = config_dict
                else:
                    configs[(config_dict["measurementType"], config_dict["technologyType"])] = config_dict
        except ValidationError:
            errors.append({
                "message": "Configurations could not be loaded",
                "supplemental": "On loading {}".format(file),
                "code": 4001
            })
            log.error("(E) Could not load configuration file {}".format(os.path.basename(file)))
    return configs


def check_measurement_technology_types(assay_json, configs):
    try:
        measurement_type = assay_json["measurementType"]["annotationValue"]
        technology_type = assay_json["technologyType"]["annotationValue"]
        config = configs[(measurement_type, technology_type)]
        if config is None:
            raise KeyError
    except KeyError:
        errors.append({
            "message": "Measurement/technology type invalid",
            "supplemental": "Measurement {}/technology {}".format(measurement_type, technology_type),
            "code": 4002
        })
        log.error("(E) Could not load configuration for measurement type '{}' and technology type '{}'"
                  .format(measurement_type, technology_type))


def check_study_and_assay_graphs(study_json, configs):

    def check_assay_graph(process_sequence_json, config):
        list_of_last_processes_in_sequence = [i for i in process_sequence_json if "nextProcess" not in i.keys()]
        log.info("Checking against assay protocol sequence configuration {}".format(config["description"]))
        config_protocol_sequence = [i["protocol"] for i in config["protocols"]]
        for process in list_of_last_processes_in_sequence:  # build graphs backwards
            assay_graph = list()
            try:
                while True:
                    process_graph = list()
                    if "outputs" in process.keys():
                        outputs = process["outputs"]
                        if len(outputs) > 0:
                            for output in outputs:
                                output_id = output["@id"]
                                process_graph.append(output_id)
                    protocol_id = protocols_and_types[process["executesProtocol"]["@id"]]
                    process_graph.append(protocol_id)
                    if "inputs" in process.keys():
                        inputs = process["inputs"]
                        if len(inputs) > 0:
                            for input_ in inputs:
                                input_id = input_["@id"]
                                process_graph.append(input_id)
                    process_graph.reverse()
                    assay_graph.append(process_graph)
                    process = [i for i in process_sequence_json if i["@id"] == process["previousProcess"]["@id"]][0]
                    if process['@id'] == process["previousProcess"]["@id"]:
                        log.fatal("Previous process is same as current process, which forms a loop!!!!! Cannot find start node!!!!!!!")
                        break
            except KeyError:  # this happens when we can"t find a previousProcess
                pass
            assay_graph.reverse()
            assay_protocol_sequence = [[j for j in i if not j.startswith("#")] for i in assay_graph]
            assay_protocol_sequence = [i for j in assay_protocol_sequence for i in j]  # flatten list
            assay_protocol_sequence_of_interest = [i for i in assay_protocol_sequence if i in config_protocol_sequence]
            #  filter out protocols in sequence that are not of interest (additional ones to required by config)
            squished_assay_protocol_sequence_of_interest = list()
            prev_prot = None
            for prot in assay_protocol_sequence_of_interest:  # remove consecutive same protocols
                if prev_prot != prot:
                    squished_assay_protocol_sequence_of_interest.append(prot)
                prev_prot = prot
            from isatools.utils import contains
            if not contains(squished_assay_protocol_sequence_of_interest, config_protocol_sequence):
                warnings.append({
                    "message": "Process sequence is not valid against configuration",
                    "supplemental": "Config protocol sequence {} does not in assay protocol sequence {}".format(config_protocol_sequence,
                                                                                                                squished_assay_protocol_sequence_of_interest),
                    "code": 4004
                })
                log.warning("Configuration protocol sequence {} does not match study graph found in {}"
                            .format(config_protocol_sequence, assay_protocol_sequence))

    protocols_and_types = dict([(i["@id"], i["protocolType"]["annotationValue"]) for i in study_json["protocols"]])
    # first check study graph
    log.info("Loading configuration (study)")
    config = configs["study"]
    check_assay_graph(study_json["processSequence"], config)
    for assay_json in study_json["assays"]:
        m = assay_json["measurementType"]["annotationValue"]
        t = assay_json["technologyType"]["annotationValue"]
        log.info("Loading configuration ({}, {})".format(m, t))
        config = configs[(m, t)]
        check_assay_graph(assay_json["processSequence"], config)


def check_study_groups(study_or_assay):
    samples = study_or_assay.samples
    study_groups = set()
    for sample in samples:
        if len(sample.factor_values) > 0:
            factors = tuple(sample.factor_values)
            study_groups.add(factors)
    num_study_groups = len(study_groups)
    log.info('Found {} study groups in {}'.format(num_study_groups,
                                                  study_or_assay.identifier))
    info.append({
        'message': 'Found {} study groups in {}'.format(
            num_study_groups, study_or_assay.identifier),
        'supplemental': 'Found {} study groups in {}'.format(
            num_study_groups, study_or_assay.identifier),
        'code': 5001
    })
    study_group_size_in_comment = study_or_assay.get_comment(
        'Number of Study Groups')
    if study_group_size_in_comment is not None:
        if study_group_size_in_comment != num_study_groups:
            warnings.append({
                'message': 'Reported study group size does not match table'
                    .format(num_study_groups,
                            study_or_assay.identifier),
                'supplemental': 'Study group size reported as {} but found {} '
                                'in {}'.format(
                    study_group_size_in_comment, num_study_groups,
                    study_or_assay.identifier),
                'code': 5002
            })


BASE_DIR = os.path.dirname(__file__)
default_config_dir = os.path.join(BASE_DIR, "resources", "config", "json", "default")
default_isa_json_schemas_dir = os.path.join(
    BASE_DIR,
    "resources",
    "schemas",
    "isa_model_version_1_0_schemas",
    "core"
)


def validate(
        fp,
        config_dir=default_config_dir,
        log_level=None,
        base_schemas_dir="isa_model_version_1_0_schemas"
):
    if config_dir is None:
        config_dir = default_config_dir
    if log_level in (
        logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING,
        logging.ERROR, logging.CRITICAL):
        log.setLevel(log_level)
    log.info("ISA JSON Validator from ISA tools API v0.12.")
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    log.addHandler(handler)
    try:
        global errors
        errors = list()
        global warnings
        warnings = list()
        log.info("Checking if encoding is UTF8")
        check_utf8(fp=fp)  # Rule 0010
        log.info("Loading json from " + fp.name)
        isa_json = json.load(fp=fp)  # Rule 0002
        log.info("Validating JSON against schemas using Draft4Validator")
        check_isa_schemas(isa_json=isa_json,
                          investigation_schema_path=os.path.join(BASE_DIR, "resources", "schemas", base_schemas_dir,
                                                                 "core", "investigation_schema.json"))  # Rule 0003
        log.info("Checking if material IDs used are declared...")
        for study_json in isa_json["studies"]:
            check_material_ids_not_declared_used(study_json)  # Rules 1002-1005
        for study_json in isa_json["studies"]:
            check_material_ids_declared_used(study_json, get_source_ids)  # Rule 1015
            check_material_ids_declared_used(study_json, get_sample_ids)  # Rule 1016
            check_material_ids_declared_used(study_json, get_material_ids)  # Rule 1017
            check_material_ids_declared_used(study_json, get_data_file_ids)  # Rule 1018
        log.info("Checking characteristic categories usage...")
        check_characteristic_category_ids_usage(isa_json["studies"])  # Rules 1013 and 1022
        log.info("Checking study factor usage...")
        for study_json in isa_json["studies"]:
            check_study_factor_usage(study_json)  # Rules 1008 and 1021
        log.info("Checking protocol parameter usage...")
        for study_json in isa_json["studies"]:
            check_protocol_parameter_ids_usage(study_json)  # Rules 1009 and 1020
        log.info("Checking unit category usage...")
        for study_json in isa_json["studies"]:
            check_unit_category_ids_usage(study_json)  # Rules 1014 and 1022
        log.info("Checking process sequences (study)...")
        for study_json in isa_json["studies"]:
            check_process_sequence_links(study_json["processSequence"])  # Rule 1006
            log.info("Checking process sequences (assay)...")
            for assay_json in study_json["assays"]:
                check_process_sequence_links(assay_json["processSequence"])  # Rule 1006
        log.info("Checking process protocol usage...")
        for study_json in isa_json["studies"]:
            check_process_protocol_ids_usage(study_json)  # Rules 1007 and 1019
        log.info("Checking date formats...")
        check_date_formats(isa_json)  # Rule 3001
        log.info("Checking DOI formats...")
        check_dois(isa_json)  # Rule 3002
        log.info("Checking Pubmed ID formats...")
        check_pubmed_ids_format(isa_json)  # Rule 3003
        log.info("Checking filenames are present...")
        check_filenames_present(isa_json)  # Rule 3005
        log.info("Checking protocol names...")
        check_protocol_names(isa_json)  # Rule 1010
        log.info("Checking protocol parameter names...")
        check_protocol_parameter_names(isa_json)  # Rule 1011
        log.info("Checking study factor names...")
        check_study_factor_names(isa_json)  # Rule 1012
        log.info("Checking ontology sources...")
        check_ontology_sources(isa_json)  # Rule 3008
        log.info("Checking term source REFs...")
        check_term_source_refs(isa_json)  # Rules 3007 and 3009
        log.info("Checking missing term source REFs...")
        check_term_accession_used_no_source_ref(isa_json)  # Rule 3010
        log.info("Loading configurations from " + config_dir)
        configs = load_config(config_dir)  # Rule 4001
        log.info("Checking measurement and technology types...")
        for study_json in isa_json["studies"]:
            for assay_json in study_json["assays"]:
                check_measurement_technology_types(assay_json, configs)  # Rule 4002
        log.info("Checking against configuration schemas...")
        check_isa_schemas(
            isa_json=isa_json,
            investigation_schema_path=os.path.join(
                default_isa_json_schemas_dir,
                "investigation_schema.json"
            )
        )  # Rule 4003
        # if all ERRORS are resolved, then try and validate against configuration
        handler.flush()
        if "(E)" in stream.getvalue():
            log.fatal("(F) There are some errors that mean validation against configurations cannot proceed.")
            return stream
        fp.seek(0)  # reset file pointer
        log.info("Checking study and assay graphs...")
        for study_json in isa_json["studies"]:
            check_study_and_assay_graphs(study_json, configs)  # Rule 4004
        fp.seek(0)
        # try load and do study groups check
        log.info("Checking study groups...")
        isa = load(fp)
        for study in isa.studies:
            check_study_groups(study)
            for assay in study.assays:
                check_study_groups(assay)
        log.info("Finished validation...")
    except KeyError as k:
        errors.append({
            "message": "JSON Error",
            "supplemental": "Error when reading JSON; key: {}".format(str(k)),
            "code": 2
        })
        log.fatal("(F) There was an error when trying to read the JSON")
        log.fatal("Key: " + str(k))
    except ValueError as v:
        errors.append({
            "message": "JSON Error",
            "supplemental": "Error when parsing JSON; key: {}".format(str(v)),
            "code": 2
        })
        log.fatal("(F) There was an error when trying to parse the JSON")
        log.fatal("Value: " + str(v))
    except SystemError as e:
        errors.append({
            "message": "Unknown/System Error",
            "supplemental": str(e),
            "code": 0
        })
        log.fatal("(F) Something went very very wrong! :(")
    finally:
        handler.flush()
        return {
            "errors": errors,
            "warnings": warnings,
            "validation_finished": True
        }


def batch_validate(json_file_list):
    """ Validate a batch of ISA-JSON files
        :param json_file_list: List of file paths to the ISA-JSON files to validate
        :return: Dict of reports

        Example:
            from isatools import isajson
            my_jsons = [
                "/path/to/study1.json",
                "/path/to/study2.json"
            ]
            my_reports = isajson.batch_validate(my_jsons)
        """
    batch_report = {
        "batch_report": []
    }
    for json_file in json_file_list:
        log.info("***Validating {}***\n".format(json_file))
        if not os.path.isfile(json_file):
            log.warning("Could not find ISA-JSON file, skipping {}".format(json_file))
        else:
            with open(json_file) as fp:
                batch_report["batch_report"].append(
                    {
                        "filename": fp.name,
                        "report": validate(fp)
                    }
                )
    return batch_report


class ISAJSONEncoder(JSONEncoder):

    def default(self, o):

        def remove_nulls(d):
            return {k: v for k, v in d.items() if v or isinstance(v, list) or v == ''}

        def nulls_to_str(d):
            to_del = []
            for k, v in d.items():
                if not isinstance(v, list) and v is None:
                    d[k] = ''
                if (k == "unit" or k == "previousProcess" or k == "nextProcess") and (v is None or v == ''):
                    to_del.append(k)
            for k in to_del:
                del d[k]
            return d

        # TODO: deal with non-verbose mode parsing; currently will break because of missing k-v's
        clean_nulls = nulls_to_str  # creates verbose JSON if using nulls to str
        # clean_nulls = remove_nulls  # optimises by removing k-v's that are null or empty strings but breaks reader

        def get_comment(obj):
            return clean_nulls(
                {
                    "name": obj.name,
                    "value": obj.value
                }
            )

        def get_comments(obj):
            return list(map(lambda x: get_comment(x), obj if obj else []))

        def get_ontology_source(obj):
            return clean_nulls(
                {
                    "name": obj.name,
                    "description": obj.description,
                    "file": obj.file,
                    "version": obj.version,
                    "comments": get_comments(obj.comments) if obj.comments else []
                }
            )

        def get_ontology_annotation(obj):
            if obj is not None:
                return clean_nulls(
                    {
                        "@id": id_gen(obj),
                        "annotationValue": obj.term,
                        "termAccession": obj.term_accession,
                        "termSource": obj.term_source.name if obj.term_source else None,
                        "comments": get_comments(obj.comments)
                    }
                )
            else:
                return None

        def get_ontology_annotations(obj):
            return list(map(lambda x: get_ontology_annotation(x), obj))

        def get_person(obj):
            return clean_nulls(
                {
                    "address": obj.address,
                    "affiliation": obj.affiliation,
                    "comments": get_comments(obj.comments),
                    "email": obj.email,
                    "fax": obj.fax,
                    "firstName": obj.first_name,
                    "lastName": obj.last_name,
                    "midInitials": obj.mid_initials if obj.mid_initials else '',
                    "phone": obj.phone,
                    "roles": get_ontology_annotations(obj.roles)
                }
            )

        def get_people(obj):
            return list(map(lambda x: get_person(x), obj))

        def get_publication(obj):
            return clean_nulls(
                {
                    "authorList": obj.author_list,
                    "doi": obj.doi,
                    "pubMedID": obj.pubmed_id,
                    "status": get_ontology_annotation(obj.status) if obj.status else {"@id": ''},
                    "title": obj.title,
                    "comments": get_comments(obj.comments) if obj.comments else []
                }
            )

        def get_publications(obj):
            return list(map(lambda x: get_publication(x), obj))

        def get_protocol(obj):
            return clean_nulls(
                {
                    "@id": id_gen(obj),
                    "description": obj.description,
                    "parameters": list(map(lambda x: {
                        "@id": id_gen(x),
                        "parameterName": get_ontology_annotation(x.parameter_name)
                    }, obj.parameters)),  # TODO: Deal with Array Design REF
                    "name": obj.name,
                    "protocolType": get_ontology_annotation(obj.protocol_type),
                    "uri": obj.uri,
                    "comments": get_comments(obj.comments) if obj.comments else [],
                    "components": [],  # TODO: Output components
                    "version": obj.version
                }
            )

        def get_source(obj):
            return clean_nulls(
                {
                    "@id": id_gen(obj),
                    "name": obj.name,
                    "characteristics": get_characteristics(obj.characteristics),
                    "comments": get_comments(obj.comments) if obj.comments else []
                }
            )

        def get_characteristic(obj):
            return clean_nulls(
                {
                    "category": {"@id": id_gen(obj.category)} if obj.category else None,
                    # "category": get_value(o.category) if o.category else None,
                    "value": get_value(obj.value),
                    "unit": {"@id": id_gen(obj.unit)} if obj.unit else None,
                    "comments": get_comments(obj.comments) if obj.comments else []
                }
            )

        def get_characteristics(obj):
            return list(map(lambda x: get_characteristic(x), obj))

        def get_value(obj):
            if isinstance(obj, OntologyAnnotation):
                return get_ontology_annotation(obj)
            elif isinstance(obj, (str, int, float)):
                return obj
            else:
                raise ValueError("Unexpected value type found: " + type(obj))

        def get_characteristic_category(obj):  # TODO: Deal with Material Type
            res = clean_nulls(
                {
                    "@id": id_gen(obj),
                    "characteristicType": get_ontology_annotation(obj)
                }
            )
            return res

        def get_sample(obj):
            return clean_nulls(
                {
                    "@id": id_gen(obj),
                    "name": obj.name,
                    "characteristics": get_characteristics(obj.characteristics),
                    "factorValues": list(map(lambda x: clean_nulls(
                        {
                            "category": {"@id": id_gen(x.factor_name)} if x.factor_name else None,
                            "value": get_value(x.value),
                            "unit": {"@id": id_gen(x.unit)} if x.unit else None
                        }
                    ), obj.factor_values)),
                    "comments": get_comments(obj.comments) if obj.comments else []
            })

        def get_factor(obj):
            return clean_nulls(
                {
                    "@id": id_gen(obj),
                    "factorName": obj.name,
                    "factorType": get_ontology_annotation(obj.factor_type),
                    "comments": get_comments(obj.comments) if obj.comments else []
                }
            )

        def get_other_material(obj):
            return clean_nulls(
                {
                    "@id": id_gen(obj),
                    "name": obj.name,
                    "type": obj.type,
                    "characteristics": get_characteristics(obj.characteristics),
                    "comments": get_comments(obj.comments) if obj.comments else []
                }
            )

        def sqeezstr(s):
            return s.replace(' ', '').lower()

        def id_gen(obj):
            if obj is not None:
                o_id = getattr(obj, 'id', None)
                if not o_id:
                    o_id = str(id(obj))
                if isinstance(obj, Source):
                    return '#source/' + o_id
                elif isinstance(obj, Sample):
                    return '#sample/' + o_id
                elif isinstance(obj, Material):
                    if obj.type == 'Extract Name':
                        return '#material/extract-' + o_id
                    elif obj.type == 'Labeled Extract Name':
                        return '#material/labledextract-' + o_id
                    else:
                        raise TypeError("Could not resolve data type labeled: " + obj.type)
                elif isinstance(obj, DataFile):
                        return '#data/{}-'.format(sqeezstr(obj.label)) + o_id
                elif isinstance(obj, Process):
                    return '#process/' + o_id  # TODO: Implement ID gen on different kinds of processes?
                else:
                    return '#' + o_id
            else:
                return None

        def get_process(obj):
            return clean_nulls(
                {
                    "@id": id_gen(obj),
                    "name": obj.name,
                    "executesProtocol": {"@id": id_gen(obj.executes_protocol)},
                    "parameterValues": list(map(lambda x: get_parameter_value(x), obj.parameter_values)),
                    "performer": obj.performer,
                    "date": obj.date,
                    "previousProcess": {"@id": id_gen(obj.prev_process)} if obj.prev_process else None,
                    "nextProcess": {"@id": id_gen(obj.next_process)} if obj.next_process else None,
                    "inputs": list(map(lambda x: {"@id": id_gen(x)}, obj.inputs)),
                    "outputs": list(map(lambda x: {"@id": id_gen(x)}, obj.outputs)),
                    "comments": get_comments(obj.comments)
                }
            )

        def get_parameter_value(obj):
            return clean_nulls(
                {
                    "category": {"@id": id_gen(obj.category)} if obj.category else None,
                    "value": get_value(obj.value),
                    "unit": {"@id": id_gen(obj.unit)} if obj.unit else None
                }
            )

        def get_study(obj): return clean_nulls(
            {
                "filename": obj.filename,
                "identifier": obj.identifier,
                "title": obj.title,
                "description": obj.description,
                "submissionDate": obj.submission_date,
                "publicReleaseDate": obj.public_release_date,
                "publications": get_publications(obj.publications),
                "people": get_people(obj.contacts),
                "studyDesignDescriptors": get_ontology_annotations(obj.design_descriptors),
                "protocols": list(map(lambda x: get_protocol(x), obj.protocols)),
                "materials": {
                    "sources": list(map(lambda x: get_source(x), obj.sources)),
                    "samples": get_samples(obj.samples),
                    "otherMaterials": get_other_materials(obj.other_material)
                },
                "processSequence": list(map(lambda x: get_process(x), obj.process_sequence)),
                "factors": list(map(lambda x: get_factor(x), obj.factors)),
                "characteristicCategories": get_characteristic_categories(obj.characteristic_categories),
                "unitCategories": get_ontology_annotations(obj.units),
                "comments": get_comments(obj.comments),
                "assays": list(map(lambda x: get_assay(x), obj.assays))
            }
        )

        def get_characteristic_categories(obj):
            return list(map(lambda x: get_characteristic_category(x), obj))

        def get_samples(obj):
            return list(map(lambda x: get_sample(x), obj))

        def get_other_materials(obj):
            return list(map(lambda x: get_other_material(x), obj))

        def get_processes(obj):
            return list(map(lambda x: get_process(x), obj))

        def get_assay(obj):
            return clean_nulls(
                {
                    "measurementType": get_ontology_annotation(obj.measurement_type),
                    "technologyType": get_ontology_annotation(obj.technology_type),
                    "technologyPlatform": obj.technology_platform,
                    "filename": obj.filename,
                    "characteristicCategories": get_characteristic_categories(obj.characteristic_categories),
                    "unitCategories": get_ontology_annotations(obj.units),
                    "comments": get_comments(obj.comments) if obj.comments else [],
                    "materials": {
                        "samples": get_samples(obj.samples),
                        "otherMaterials": get_other_materials(obj.other_material)
                    },
                    "dataFiles": list(map(lambda x: get_data_file(x), obj.data_files)),
                    "processSequence": get_processes(obj.process_sequence)
                }
            )

        def get_data_file(obj):
            return clean_nulls(
                {
                    "@id": id_gen(obj),
                    "name": obj.filename,
                    "type": obj.label,
                    "comments": get_comments(obj.comments)
                }
            )

        if isinstance(o, Investigation):
            return clean_nulls(
                {
                    "identifier": o.identifier,
                    "title": o.title,
                    "description": o.description,
                    "comments": get_comments(o.comments),
                    "ontologySourceReferences": list(map(lambda x: get_ontology_source(x), o.ontology_source_references)),
                    "people": get_people(o.contacts),
                    "publicReleaseDate": o.public_release_date,
                    "submissionDate": o.submission_date,
                    "publications": get_publications(o.publications),
                    "studies": list(map(lambda x: get_study(x), o.studies))
                }
            )
        elif isinstance(o, Study):
            return get_study(o)
        elif isinstance(o, OntologySource):
            return get_ontology_source(o)
        elif isinstance(o, OntologyAnnotation):
            return get_ontology_annotation(o)
        elif isinstance(o, Person):
            return get_person(o)
        elif isinstance(o, Publication):
            return get_publication(o)
        elif isinstance(o, Protocol):
            return get_protocol(o)
        elif isinstance(o, Characteristic):
            return get_characteristic(o)
        # TODO: enable dump of all objects and add some tests on them


# -*- coding: utf-8 -*
"""Convert MAGE-TAB to ISA-JSON"""
from __future__ import absolute_import
import json
import os
import shutil
import tempfile

from isatools import isatab
from isatools.convert import magetab2isatab
from isatools.isajson import ISAJSONEncoder


def convert(idf_file_path):
    tmp = tempfile.mkdtemp()
    ISA = None
    try:
        magetab2isatab.convert(idf_file_path, output_path=tmp)
        with open(os.path.join(
                tmp, 'i_investigation.txt'), encoding='utf-8') as isa_inv_fp:
            ISA = isatab.load(isa_inv_fp)
    finally:
        shutil.rmtree(tmp)
        if ISA is not None:
            return json.loads(json.dumps(ISA, cls=ISAJSONEncoder))



# -*- coding: utf-8 -*-
"""Functions for reading and writing MAGE-TAB.

Functions for reading and writing MAGE-TAB. MAGE-TAB content is loaded into an
in-memory representation using the ISA Data Model implemented in the
isatools.model package.
"""
from __future__ import absolute_import
import copy
import csv
import os
import re
import tempfile
from io import StringIO
from itertools import zip_longest
import numpy as np
import pandas as pd
from isatools import isatab
from isatools.model import (
    Assay,
    Comment,
    Investigation,
    OntologyAnnotation,
    OntologySource,
    Person,
    Protocol,
    ProtocolParameter,
    Publication,
    Study,
    StudyFactor,
)
import logging

log = logging.getLogger('isatools')


def _get_sdrf_filenames(ISA):
    sdrf_filenames = []
    for study in ISA.studies:
        for assay in [x for x in study.assays if
                      x.technology_type.term.lower() == "dna microarray"]:
            sdrf_filenames.append(study.filename[2:-3] +
                                  assay.filename[2:-3] + "sdrf.txt")
    return sdrf_filenames


def _build_metadata_df(ISA):
    metadata_df = pd.DataFrame(columns=(
        "MAGE-TAB Version",
        "Investigation Title",
        "Public Release Date",
        "SDRF File"
    ))
    sdrf_filenames = _get_sdrf_filenames(ISA)
    metadata_df.loc[0] = [
        "1.1",
        ISA.title,
        ISA.public_release_date,
        sdrf_filenames[0]
    ]
    for i, sdrf_filename in enumerate(sdrf_filenames):
        if i == 0:
            pass
        else:
            metadata_df.loc[i] = [
                "",
                "",
                "",
                sdrf_filename
            ]
    return metadata_df


def _build_exp_designs_df(ISA):
    exp_designs_df = pd.DataFrame(columns=(
        "Experimental Design",
        "Experimental Design Term Source REF",
        "Experimental Design Term Accession Number"))
    microarray_study_design = []
    for study in ISA.studies:
        if len([x for x in study.assays
                if x.technology_type.term.lower() == "dna microarray"]) > 0:
            microarray_study_design.extend(study.design_descriptors)
    for i, design_descriptor in enumerate(microarray_study_design):
        exp_designs_df.loc[i] = [
            design_descriptor.term,
            design_descriptor.term_source.name,
            design_descriptor.term_accession
        ]
    return exp_designs_df


def _build_exp_factors_df(ISA):
    exp_factors_df = pd.DataFrame(columns=(
        "Experimental Factor Name",
        "Experimental Factor Type",
        "Experimental Factor Term Source REF",
        "Experimental Factor Term Accession Number"))
    microarray_study_factors = []
    for study in ISA.studies:
        if len([x for x in study.assays
                if x.technology_type.term.lower() == "dna microarray"]) > 0:
            microarray_study_factors.extend(study.factors)
    for i, factor in enumerate(microarray_study_factors):
        exp_factors_df.loc[i] = [
            factor.name,
            factor.factor_type.term,
            factor.factor_type.term_source.name
            if factor.factor_type.term_source else "",
            factor.factor_type.term_accession
            if factor.factor_type.term_source else ""
        ]
    return exp_factors_df


def _build_roles_str(roles):
    if roles is None:
        roles = list()
    roles_names = ''
    roles_accession_numbers = ''
    roles_source_refs = ''
    for role in roles:
        roles_names += (role.term if role.term else '') + ';'
        roles_accession_numbers += (
            role.term_accession if role.term_accession else '') + ';'
        roles_source_refs += (role.term_source.name
                              if role.term_source else '') + ';'
    if len(roles) > 0:
        roles_names = roles_names[:-1]
        roles_accession_numbers = roles_accession_numbers[:-1]
        roles_source_refs = roles_source_refs[:-1]
    return roles_names, roles_accession_numbers, roles_source_refs


def _build_people_df(ISA):
    people_df = pd.DataFrame(columns=("Person Last Name",
                                      "Person Mid Initials",
                                      "Person First Name",
                                      "Person Email",
                                      "Person Phone",
                                      "Person Fax",
                                      "Person Address",
                                      "Person Affiliation",
                                      "Person Roles",
                                      "Person Roles Term Source REF",
                                      "Person Roles Term Accession Number"))
    for i, contact in enumerate(ISA.contacts):
        roles_names, roles_accessions, roles_sources = _build_roles_str(
            contact.roles)
        people_df.loc[i] = [
            contact.last_name,
            contact.mid_initials,
            contact.first_name,
            contact.email,
            contact.phone,
            contact.fax,
            contact.address,
            contact.affiliation,
            roles_names,
            roles_sources,
            roles_accessions
        ]
    return people_df


def _build_protocols_df(ISA):
    protocols_df = pd.DataFrame(columns=('Protocol Name',
                                         'Protocol Type',
                                         'Protocol Term Accession Number',
                                         'Protocol Term Source REF',
                                         'Protocol Description',
                                         'Protocol Parameters',
                                         'Protocol Hardware',
                                         'Protocol Software',
                                         'Protocol Contact'
                                         )
                                )
    microarray_study_protocols = []
    for study in ISA.studies:
        if len([x for x in study.assays if x.technology_type.term.lower()
                == "dna microarray"]) > 0:
            microarray_study_protocols.extend(study.protocols)
    for i, protocol in enumerate(microarray_study_protocols):
        parameters_names = ''
        parameters_accession_numbers = ''
        parameters_source_refs = ''
        for parameter in protocol.parameters:
            parameters_names += parameter.parameter_name.term + ';'
            parameters_accession_numbers += (
                parameter.parameter_name.term_accession
                if parameter.parameter_name.term_accession is not None
                else '') + ';'
            parameters_source_refs += (
                parameter.parameter_name.term_source.name
                if parameter.parameter_name.term_source else '') + ';'
        if len(protocol.parameters) > 0:
            parameters_names = parameters_names[:-1]
        if protocol.protocol_type:
            protocol_type_term = protocol.protocol_type.term
            protocol_type_term_accession = \
                protocol.protocol_type.term_accession
            if protocol.protocol_type.term_source:
                protocol_type_term_source_name = \
                    protocol.protocol_type.term_source.name
                protocols_df.loc[i] = [
                    protocol.name,
                    protocol_type_term,
                    protocol_type_term_accession,
                    protocol_type_term_source_name,
                    protocol.description,
                    parameters_names,
                    "",
                    "",
                    ""
                ]
    return protocols_df


def _build_term_sources_df(ISA):
    term_sources_df = pd.DataFrame(
        columns=(
            "Term Source Name",
            "Term Source File",
            "Term Source Version"))
    for i, term_source in enumerate(ISA.ontology_source_references):
        term_sources_df.loc[i] = [
            term_source.name,
            term_source.file,
            term_source.version
        ]
    return term_sources_df


def _build_publications_df(ISA):
    publications = ISA.studies[0].publications
    publications_df_cols = ['PubMed ID',
                            'Publication DOI',
                            'Publication Author List',
                            'Publication Title',
                            'Publication Status',
                            'Publication Status Term Accession Number',
                            'Publication Status Term Source REF']
    if len(publications) > 0:
        try:
            for comment in publications[0].comments:
                publications_df_cols.append('Comment[' + comment.name + ']')
        except TypeError:
            pass
    publications_df = pd.DataFrame(columns=tuple(publications_df_cols))
    for i, publication in enumerate(publications):
        if publication.status is not None:
            status_term = publication.status.term
            status_term_accession = publication.status.term_accession
            if publication.status.term_source is not None:
                status_term_source_name = publication.status.term_source.name
            else:
                status_term_source_name = ''
        else:
            status_term = ''
            status_term_accession = ''
            status_term_source_name = ''
        publications_df_row = [
            publication.pubmed_id,
            publication.doi,
            publication.author_list,
            publication.title,
            status_term,
            status_term_accession,
            status_term_source_name,
        ]
        try:
            for comment in publication.comments:
                publications_df_row.append(comment.value)
        except TypeError:
            pass
        publications_df.loc[i] = publications_df_row
    return publications_df


def _build_qc_df(ISA):
    qc_df = pd.DataFrame(columns=(
        "Quality Control Type",
        "Quality Control Term Accession Number",
        "Quality Control Term Source REF"))
    for i, qc_comment in enumerate(
            [x for x in ISA.studies[0].comments
             if x.name == "Quality Control Type"]):
        qc_df.loc[i] = [
            qc_comment.value,
            "",
            ""
        ]
    return qc_df


def _build_replicates_df(ISA):
    replicates_df = pd.DataFrame(columns=(
        "Replicate Type",
        "Replicate Term Accession Number",
        "Replicate Term Source REF"))
    for i, replicate_comment in enumerate(
            [x for x in ISA.studies[0].comments
             if x.name == "Replicate Type"]):
        replicates_df.loc[i] = [
            replicate_comment.value,
            "",
            ""
        ]
    return replicates_df


def _build_normalizations_df(ISA):
    normalizations_df = pd.DataFrame(columns=(
        "Normalization Type",
        "Normalization Term Accession Number",
        "Normalization Term Source REF"))
    for i, normalization_comment in enumerate(
            [x for x in ISA.studies[0].comments
             if x.name == "Normalization Type"]):
        normalizations_df.loc[i] = [
            normalization_comment.value,
            "",
            ""
        ]
    return normalizations_df


def write_idf_file(inv_obj, output_path):
    """Writes IDF file out from ISA objects

    :param inv_obj: ISA Investigation object
    :param output_path: Output path to write IDF to
    :return: None
    """
    investigation = inv_obj
    metadata_df = _build_metadata_df(investigation)
    exp_designs_df = _build_exp_designs_df(investigation)
    exp_factors_df = _build_exp_factors_df(investigation)

    qc_df = _build_qc_df(investigation)
    replicates_df = _build_replicates_df(investigation)
    normalizations_df = _build_normalizations_df(investigation)

    people_df = _build_people_df(investigation)
    publications_df = _build_publications_df(investigation)
    protocols_df = _build_protocols_df(investigation)
    term_sources_df = _build_term_sources_df(investigation)

    idf_df = pd.concat([
        metadata_df,
        exp_designs_df,
        exp_factors_df,
        qc_df,
        replicates_df,
        normalizations_df,
        people_df,
        publications_df,
        protocols_df,
        term_sources_df
    ], axis=1)
    idf_df = idf_df.set_index("MAGE-TAB Version").T
    idf_df = idf_df.replace('', np.nan)
    with open(os.path.join(output_path, "{}.idf.txt".format(
            investigation.identifier if investigation.identifier != ""
            else investigation.filename[2:-3])), "w",
            encoding='utf-8') as idf_fp:
        idf_df.to_csv(
            path_or_buf=idf_fp,
            index=True,
            sep='\t',
            encoding='utf-8',
            index_label="MAGE-TAB Version")


def write_sdrf_table_files(i, output_path):
    """Writes out SDRF table files

    :param i: ISA Investigation object
    :param output_path: Output path to write SDRFs to
    :return: None
    """
    tmp = tempfile.mkdtemp()
    isatab.write_study_table_files(inv_obj=i, output_dir=tmp)
    isatab.write_assay_table_files(inv_obj=i, output_dir=tmp)
    for study in i.studies:
        for assay in [
            x for x in study.assays
                if x.technology_type.term.lower() == "dna microarray"]:
            sdrf_filename = study.filename[2:-3] + \
                assay.filename[2:-3] + "sdrf.txt"
            log.debug("Writing {}".format(sdrf_filename))
            try:
                isatab.merge_study_with_assay_tables(
                    os.path.join(tmp, study.filename),
                    os.path.join(tmp, assay.filename),
                    os.path.join(output_path, sdrf_filename))
            except FileNotFoundError:
                raise IOError("There was a problem merging intermediate "
                              "ISA-Tab files into SDRF")


def dump(inv_obj, output_path):
    """Dumps out MAGE-TAB files

    :param inv_obj: An Investigation object
    :param output_path: Target path to write IDF and SDRF MAGE-TAB files to
    :return: the origin inv_obj
    """
    num_microarray_assays = 0
    for study in inv_obj.studies:
        num_microarray_assays += len(
            [x for x in study.assays
             if x.technology_type.term.lower() == "dna microarray"])

    if num_microarray_assays > 0:
        write_idf_file(inv_obj, output_path=output_path)
        write_sdrf_table_files(i=inv_obj, output_path=output_path)
    else:
        raise IOError("Input must contain at least one assay of type DNA "
                      "microarray, halt writing MAGE-TAB")
    return inv_obj


inv_to_idf_map = {
    "Study File Name": "SDRF File",
    "Study Title": "Investigation Title",
    "Study Description": "Experiment Description",
    "Study Public Release Date": "Public Release Date",
    "Comment[MAGETAB TimeStamp_Version]": "MAGETAB TimeStamp_Version",
    "Comment[ArrayExpressReleaseDate]": "ArrayExpressReleaseDate",
    "Comment[Date of Experiment]": "Date of Experiment",
    "Comment[AEMIAMESCORE]": "AEMIAMESCORE",
    "Comment[Submitted Name]": "Submitted Name",
    "Comment[ArrayExpressAccession]": "ArrayExpressAccession",
    "Study Design Type": "Experimental Design",
    "Study Design Type Term Accession Number": "Experimental Design "
                                               "Term Accession Number",
    "Study Design Type Ter Source REF": "Experimental Design Term Source REF",
    "Study Factor Name": "Experimental Factor Name",
    "Study Factor Type": "Experimental Factor Type",
    "Study Factor Type Term Accession Number": "Experimental Factor Type "
                                               "Term Accession Number",
    "Study Factor Type Ter Source REF": "Experimental Factor Type "
                                        "Term Source REF",
    "Study PubMed ID": "PubMed ID",
    "Study Publication DOI": "Publication DOI",
    "Study Publication Author List": "Publication Author List",
    "Study Publication Title": "Publication Title",
    "Study Publication Status": "Publication Status",
    "Study Publication Status Term Accession Number": "Publication Status "
                                                      "Term Accession Number",
    "Study Publication Status Term Source REF": "Publication Status "
                                                "Term Source REF",
    "Study Person Last Name": "Person Last Name",
    "Study Person First Name": "Person First Name",
    "Study Person Mid Initials": "Person Mid Initials",
    "Study Person Email": "Person Email",
    "Study Person Phone": "Person Phone",
    "Study Person Fax": "Person Fax",
    "Study Person Address": "Person Address",
    "Study Person Affiliation": "Person Affiliation",
    "Study Person Roles": "Person Role",
    "Study Person Roles Term Accession Number": "Person Role "
                                                "Term Accession Number",
    "Study Person Roles Term Source REF": "Person Role Term Source REF",
    "Study Protocol Name": "Protocol Name",
    "Study Protocol Description": "Protocol Description",
    "Study Protocol Parameters": "Protocol Parameters",
    "Study Protocol Type": "Protocol Type",
    "Study Protocol Type Accession Number": "Protocol Term Accession Number",
    "Study Protocol Type Source REF": "Protocol Term Source REF",
    "Comment[Protocol Software]": "Protocol Software",
    "Comment[Protocol Hardware]": "Protocol Hardware",
    "Comment[Protocol Contact]": "Protocol Contact",
    "Term Source Name": "Term Source Name",
    "Term Source File": "Term Source File",
    "Term Source Version": "Term Source Version",
    "Term Source Description": "Term Source Description"
}  # Relabel these, ignore all other lines


def cast_inv_to_idf(FP):
    # Cut out relevant Study sections from Investigation file
    idf_FP = StringIO()
    for line in FP:
        if line.startswith(tuple(inv_to_idf_map.keys())
                           ) or line.startswith("Comment["):
            for k, v in inv_to_idf_map.items():
                line = line.replace(k, v)
            idf_FP.write(line)
    idf_FP.seek(0)
    idf_FP.name = FP.name
    return idf_FP


class MageTabParserException(Exception):
    pass


def squashstr(string):
    """Squashes a string by removing the spaces and lowering it"""

    nospaces = "".join(string.split())
    return nospaces.lower()


def get_squashed(key):  # for MAGE-TAB spec 2.1.7, deal with variants on labels
    if key is None:
        return ''
    try:
        if squashstr(key[:key.index('[')]) == 'comment':
            return 'comment' + key[key.index('['):]
        else:
            return squashstr(key)
    except ValueError:
        return squashstr(key)


class MageTabParser(object):
    """ The MAGE-TAB parser
    This parses MAGE-TAB IDF and SDRF files into the Python ISA model.
    It does some best-effort inferences on missing
    metadata required by ISA, but note that outputs may still be incomplete
    and flag warnings and errors in the ISA
    validators. """

    def __init__(self):
        self.ISA = Investigation(studies=[Study()])
        self._idfdict = {}
        self._ts_dict = {}

    def parse_idf(self, in_filename):
        """Parse the MAGE-TAB IDF file

        :param in_filename: Path to the IDF file
        :return: An Investigation object with ISA content
        """
        self.load_into_idfdict(in_filename=in_filename)
        # Parse the ontology sources first, as we need to reference these later
        self.parse_ontology_sources(self._idfdict.get('termsourcename', []),
                                    self._idfdict.get('termsourcefile', []),
                                    self._idfdict.get('termsourceversion', []))
        # Then parse the rest of the sections in blocks; follows order of
        # MAGE-TAB v1.1 2011-07-28 specification
        self.parse_investigation(
            self._idfdict.get('investigationtitle', []),
            self._idfdict.get('investigationaccession', []),
            self._idfdict.get('investigationaccessiontermsourceref', []))
        self.parse_experimental_designs(
            self._idfdict.get('experimentaldesign', []),
            self._idfdict.get('experimentaldesigntermsourceref', []),
            self._idfdict.get('experimentaldesigntermaccessionnumber', []))
        self.parse_experimental_factors(
            self._idfdict.get('experimentalfactorname', []),
            self._idfdict.get('experimentalfactortype', []),
            self._idfdict.get(
                'experimentalfactortypetermsourceref', []),
            self._idfdict.get('experimentalfactortypetermaccessionnumber', []))
        self.parse_people(
            self._idfdict.get('personlastname', []),
            self._idfdict.get('personfirstname', []),
            self._idfdict.get('personmidinitials', []),
            self._idfdict.get('personemail', []),
            self._idfdict.get('personphone', []),
            self._idfdict.get('personfax', []),
            self._idfdict.get('personaddress', []),
            self._idfdict.get('personaffiliation', []),
            self._idfdict.get('personroles', []),
            self._idfdict.get('personrolestermsourceref', []),
            self._idfdict.get('personrolestermaccessionnumber', []))
        self.parse_dates(
            self._idfdict.get(
                'dateofexperiment', []), self._idfdict.get(
                'publicreleasedate', []))
        self.parse_publications(
            self._idfdict.get('pubmedid', []),
            self._idfdict.get('publicationdoi', []),
            self._idfdict.get('publicationauthorlist', []),
            self._idfdict.get('publicationtitle', []),
            self._idfdict.get('publicationstatus', []),
            self._idfdict.get(
                'publicationstatustermsourceref', []),
            self._idfdict.get('publicationstatustermaccessionnumber', []))
        self.parse_experiment_description(
            self._idfdict.get('experimentdescription'))
        self.parse_protocols(self._idfdict.get('protocolname', []),
                             self._idfdict.get('protocoltype', []),
                             self._idfdict.get('protocoltermsourceref', []),
                             self._idfdict.get(
            'protocoltermaccessionnumber', []),
            self._idfdict.get('protocoldescription', []),
            self._idfdict.get('protocolparameters', []),
            self._idfdict.get('protocolhardware', []),
            self._idfdict.get('protocolsoftware', []),
            self._idfdict.get('protocolcontact', []))
        self.parse_sdrf_file(self._idfdict.get('sdrffile', []))
        self.parse_comments(
            {key: self._idfdict[key] for key in [
                x for x in self._idfdict.keys() if x.startswith('comment[')]})
        self.infer_missing_metadata()
        return self.ISA

    def load_into_idfdict(self, in_filename):
        try:
            with open(in_filename, encoding='utf-8') as unicode_file:
                tabreader = csv.reader(
                    filter(
                        lambda r: r[0] != '#',
                        unicode_file),
                    dialect='excel-tab')
                for row in tabreader:
                    key = get_squashed(key=row[0])
                    self._idfdict[key] = row[1:]
        except UnicodeDecodeError:
            with open(in_filename, encoding='ISO8859-2') as latin2_file:
                tabreader = csv.reader(
                    filter(
                        lambda r: r[0] != '#',
                        latin2_file),
                    dialect='excel-tab')
                for row in tabreader:
                    key = get_squashed(key=row[0])
                    self._idfdict[key] = row[1:]

    def parse_ontology_sources(self, names, files, versions):
        for name, file, version in zip_longest(
                names, files, versions, fillvalue=''):
            # only add if the OS has a name and therefore can be referenced
            if name != '':
                os = OntologySource(name=name, file=file, version=version)
                self.ISA.ontology_source_references.append(os)
                self._ts_dict[name] = os

    def parse_investigation(self, titles, accessions, accessiontsrs):
        for title, accession, accessiontsr in zip_longest(
                titles, accessions, accessiontsrs, fillvalue=''):
            self.ISA.identifier = accession
            self.ISA.title = title
            self.ISA.studies[-1].title = title
            self.ISA.studies[-1].identifier = accession
            if accessiontsr is not None:
                self.ISA.comments.append(
                    Comment(
                        name="Investigation Accession Term Source REF",
                        value=accessiontsr))
            break  # because there should only be one or zero rows

    def parse_experimental_designs(self, designs, tsrs, tans):
        for design, tsr, tan in zip_longest(designs, tsrs, tans, fillvalue=''):
            design_descriptor = OntologyAnnotation(
                term=design, term_source=self._ts_dict.get(tsr),
                term_accession=tan)
            if design_descriptor.term != '':  # only add if the DD has a term
                self.ISA.studies[-1].design_descriptors.append(
                    design_descriptor)

    def parse_experimental_factors(self, factors, factortypes, tsrs, tans):
        for factor, factortype, tsr, tan in zip_longest(
                factors, factortypes, tsrs, tans, fillvalue=''):
            if factor != '':  # only add if there's a factor name
                factortype_oa = OntologyAnnotation(
                    term=factortype, term_source=self._ts_dict.get(tsr),
                    term_accession=tan)
                study_factor = StudyFactor(
                    name=factor, factor_type=factortype_oa)
                self.ISA.studies[-1].factors.append(study_factor)

    def parse_people(self, lastnames, firstnames, midinitialss, emails,
                     phones, faxes, addresses, affiliations, roles,
                     roletans, roletrs):
        for lastname, firstname, midinitials, email, phone, fax, address, \
            affiliation, role, roletan, roletsr in \
                zip_longest(lastnames, firstnames, midinitialss, emails,
                            phones, faxes, addresses, affiliations, roles,
                            roletans, roletrs, fillvalue=''):
            rolesoa = OntologyAnnotation(
                term=role,
                term_source=self._ts_dict.get(roletsr),
                term_accession=roletan)
            person = Person(last_name=lastname, first_name=firstname,
                            mid_initials=midinitials, email=email,
                            phone=phone, fax=fax, address=address,
                            affiliation=affiliation, roles=[rolesoa])
            self.ISA.studies[-1].contacts.append(person)

    def parse_dates(self, dateofexperiments, publicreleasedates):
        for dateofexperiment, publicreleasedate in zip_longest(
                dateofexperiments, publicreleasedates, fillvalue=''):
            self.ISA.public_release_date = publicreleasedate
            self.ISA.studies[-1].public_release_date = publicreleasedate
            self.ISA.studies[-1].comments.append(
                Comment(name="Date of Experiment", value=dateofexperiment))
            break  # because there should only be one or zero rows

    def parse_publications(self, pubmedids, dois, authorlists,
                           titles, statuses, statustans, statustsrs):
        for pubmedid, doi, authorlist, title, status, statustsr, statustan in \
                zip_longest(pubmedids, dois, authorlists, titles, statuses,
                            statustans, statustsrs, fillvalue=''):
            # only add if there's a pubmed ID, DOI or title
            if pubmedid != '' or doi != '' or title != '':
                statusoa = OntologyAnnotation(
                    term=status,
                    term_source=self._ts_dict.get(statustsr),
                    term_accession=statustan)
                publication = Publication(
                    pubmed_id=pubmedid,
                    doi=doi,
                    author_list=authorlist,
                    title=title,
                    status=statusoa)
                self.ISA.studies[-1].publications.append(publication)

    def parse_experiment_description(self, descriptions):
        log.info('Descriptions are: {}'.format(descriptions))
        for description in zip_longest(descriptions, fillvalue=''):
            self.ISA.studies[-1].description = description[-1]
            break  # because there should only be one or zero rows

    def parse_protocols(self, names, ptypes, tsrs, tans, descriptions,
                        parameterslists, hardwares, softwares, contacts):
        for name, ptype, tsr, tan, description, parameterslist, hardware, \
            software, contact in \
                zip_longest(names, ptypes, tsrs, tans, descriptions,
                            parameterslists, hardwares, softwares, contacts,
                            fillvalue=''):
            if name != '':  # only add if there's a name
                protocoltype_oa = OntologyAnnotation(
                    term=ptype, term_source=self._ts_dict.get(tsr),
                    term_accession=tan)
                protocol = Protocol(name=name, protocol_type=protocoltype_oa,
                                    description=description,
                                    parameters=list(map(
                                        lambda x: ProtocolParameter(
                                            parameter_name=OntologyAnnotation(
                                                term=x)),
                                        parameterslist.split(';')
                                        if parameterslist is not None
                                        else '')))
                protocol.comments = [Comment(name="Protocol Hardware",
                                             value=hardware),
                                     Comment(
                    name="Protocol Software", value=software),
                    Comment(name="Protocol Contact", value=contact)]
                self.ISA.studies[-1].protocols.append(protocol)

    def parse_sdrf_file(self, sdrffiles):
        """Parses a list of MAGE-TAB SDRF files

        :param sdrffiles:  List of SDRF files to parse
        :return: None
        """
        sdrffiles_no_empty = [x for x in sdrffiles if x != '']
        if len(sdrffiles_no_empty) > 0:
            if len(sdrffiles_no_empty) > 1:
                self.ISA.studies[-1].comments.append(
                    Comment(name="SDRF File", value=';'.join(
                        sdrffiles_no_empty)))
            else:
                self.ISA.studies[-1].comments.append(
                    Comment(name="SDRF File", value=sdrffiles_no_empty[0]))

    def parse_comments(self, commentsdict):
        for k, v in commentsdict.items():
            v_no_empty = [x for x in v if x != '']
            if len(v_no_empty) > 0:
                if len(v_no_empty) > 1:
                    v = ';'.join(v_no_empty)
                else:
                    v = v_no_empty[0]
                self.ISA.studies[-1].comments.append(
                    Comment(name=k[8:-1], value=v))

    def infer_missing_metadata(self):
        S = self.ISA.studies[-1]

        defaultassay = None
        # first let's try and infer the MT/TT from the study design
        # descriptors, only checks first one
        if len(S.design_descriptors) > 0:
            defaultassay = self._get_measurement_and_tech(
                S.design_descriptors[0].term)

        # next, go through the loaded comments to see what we can find
        for comment in S.comments:
            commentkey = get_squashed(comment.name)
            # ArrayExpress specific comments
            # (1) if there is no default assay yet, try use AEExperimentType
            if commentkey == 'aeexperimenttype' and defaultassay is None:
                defaultassay = self._get_measurement_and_tech(comment.value)
            # (2) if there is no identifier set, try use ArrayExpressAccession
            if commentkey == 'arrayexpressaccession':
                if self.ISA.identifier == '':
                    self.ISA.identifier = comment.value
                if S.identifier == '':
                    S.identifier = comment.value
            # (3) if there is no submission date set, try use
            # ArrayExpressSubmissionDate
            if commentkey == 'arrayexpresssubmissiondate':
                if self.ISA.submission_date == '':
                    self.ISA.submission_date = comment.value
                if S.submission_date == '':
                    S.submission_date = comment.value

        # if there is STILL no defaultassay set, try infer from study title
        if defaultassay is None \
                and ('transcriptionprof' in get_squashed(S.title)
                     or 'geneexpressionprof' in get_squashed(S.title)):
            defaultassay = Assay(measurement_type=OntologyAnnotation(
                term='transcription profiling'),
                technology_type=OntologyAnnotation(
                term='DNA microarray'),
                technology_platform='GeneChip')

        if defaultassay is None:
            defaultassay = Assay()

        # set file names if identifiers are available
        self.ISA.filename = 'i_{0}investigation.txt'.format(
            self.ISA.identifier + '_'
            if self.ISA.identifier != '' else self.ISA.identifier)
        S.filename = 's_{0}study.txt'.format(
            S.identifier + '_' if S.identifier != '' else S.identifier)
        defaultassay.filename = 'a_{0}assay.txt'.format(
            S.identifier + '_' if S.identifier != '' else S.identifier)

        S.assays = [defaultassay]

    @staticmethod
    def _get_measurement_and_tech(design_type):
        assay = None
        if re.match('(?i).*ChIP-Chip.*', design_type):
            assay = Assay(measurement_type=OntologyAnnotation(
                term='protein-DNA binding site identification'),
                technology_type=OntologyAnnotation(
                term='DNA microarray'),
                technology_platform='ChIP-Chip')
        if re.match('(?i).*RNA-seq.*', design_type) \
                or re.match('(?i).*RNA-Seq.*', design_type) \
                or re.match(
                '(?i).*transcription profiling by high throughput '
                'sequencing.*', design_type):
            assay = Assay(measurement_type=OntologyAnnotation(
                term='transcription profiling'),
                technology_type=OntologyAnnotation(
                term='nucleotide sequencing'),
                technology_platform='RNA-Seq')
        if re.match('.*transcription profiling by array.*', design_type) \
                or re.match('dye_swap_design', design_type):
            assay = Assay(measurement_type=OntologyAnnotation(
                term='transcription profiling'),
                technology_type=OntologyAnnotation(
                term='DNA microarray'),
                technology_platform='GeneChip')
        if re.match('(?i).*methylation profiling by array.*', design_type):
            assay = Assay(measurement_type=OntologyAnnotation(
                term='DNA methylation profiling'),
                technology_type=OntologyAnnotation(
                term='DNA microarray'),
                technology_platform='Me-Chip')
        if re.match('(?i).*comparative genomic hybridization by array.*',
                    design_type):
            assay = Assay(measurement_type=OntologyAnnotation(
                term='comparative genomic hybridization'),
                technology_type=OntologyAnnotation(
                term='DNA microarray'),
                technology_platform='CGH-Chip')
        if re.match('.*genotyping by array.*', design_type):
            assay = Assay(measurement_type=OntologyAnnotation(
                term='SNP analysis'),
                technology_type=OntologyAnnotation(
                term='DNA microarray'),
                technology_platform='SNPChip')
        if re.match('(?i).*ChIP-Seq.*',
                    design_type) or re.match('(?i).*chip-seq.*', design_type):
            assay = Assay(measurement_type=OntologyAnnotation(
                term='protein-DNA binding site identification'),
                technology_type=OntologyAnnotation(
                term='nucleotide sequencing'),
                technology_platform='ChIP-Seq')
        if assay is not None:
            assay._design_type = design_type
        return assay

    def parse_sdrf_to_isa_table_files(self, in_filename):
        """
        Parses MAGE-TAB SDRF file into ISA-Tab study and assay tables as
        pandas dataframes
        """
        with open(in_filename, encoding='utf-8') as in_fp:
            with strip_comments(in_fp) as fp:
                df = pd.read_csv(
                    fp, dtype=str, sep='\t', encoding='utf-8').fillna('')
            # do some preliminary cleanup of the table
            columns_to_keep = []
            for i, col in enumerate(df.columns):
                if col.lower().startswith('term source ref') \
                    and df.columns[i - 1].lower().startswith(
                        'protocol ref'):
                    pass
                    # drop term source ref column that appears
                    # after protocol ref
                elif col.lower().startswith('term source ref') \
                    and df.columns[i - 1].lower().startswith(
                        'array design ref'):
                    pass
                    # drop term source ref column that appears after
                    # array design ref
                elif col.lower().startswith('technology type'):
                    pass
                    # drop technology type column / in java code it moves it
                    # 1 to the right of assay name column
                elif col.lower().startswith('provider'):
                    pass  # drop provider column
                else:
                    columns_to_keep.append(col)
            # reset dataframe with only columns we are interested in
            df = df[columns_to_keep]
            #  TODO: Do we need to replicate what CleanupRunner.java does?

            # now find the first index to split the SDRF into sfile and
            # afile(s)
            cols = [x.lower() for x in list(df.columns)]  # columns all lowered
            if 'sample name' not in cols:
                # if we can't find the sample name, we need to insert it
                # somewhere
                first_node_index = -1
                if 'extract name' in cols:
                    first_node_index = cols.index('extract name')
                elif 'labeled extract name' in cols:
                    first_node_index = cols.index('labeled extract name')
                elif 'labeled extract name' in cols:
                    first_node_index = cols.index('hybridization name')
                if first_node_index > 0:  # do Sample Name insertion here
                    cols_ = list(df.columns)
                    # add Sample Name column where first indexed col is
                    df["Sample Name"] = df[cols_[first_node_index]]
                    # insert to the column index where Sample Name should occur
                    cols_.insert(first_node_index, "Sample Name")
                    # reset the dataframe with the new order of columns
                    df = df[cols_]

            # before splitting, let's rename columns where necessary

            df = df.rename(columns={
                "Material Type": "Characteristic[material]",
                "Technology Type": "Comment[technology type]",
                "Hybridization Name": "Hybridization Assay Name"
            })

            # now do the slice
            cols = list(df.columns)
            sample_name_index = cols.index("Sample Name")
            study_df = df[df.columns[0:sample_name_index + 1]
                          ].drop_duplicates()
            assay_df = df[df.columns[sample_name_index:]]

            table_files = []
            with StringIO() as assay_fp:
                columns = [x[:x.rindex('.')] if '.' in x else x for x in list(
                    assay_df.columns)]
                assay_df.columns = columns
                assay_df.to_csv(
                    path_or_buf=assay_fp,
                    mode='a',
                    sep='\t',
                    encoding='utf-8',
                    index=False)
                log.info(
                    "Trying to split assay file extracted from %s",
                    in_filename)
                assay_fp.seek(0)
                assay_files = self.split_assay(assay_fp)
                log.info("We have %s assays", len(assay_files))

            study_fp = StringIO()
            study_fp.name = self.ISA.studies[-1].filename
            columns = [x[:x.rindex('.')] if '.' in x else x for x in list(
                study_df.columns)]
            study_df.columns = columns
            study_df.to_csv(
                path_or_buf=study_fp,
                mode='a',
                sep='\t',
                encoding='utf-8',
                index=False)
            study_fp.seek(0)
            table_files.append(study_fp)
            table_files.extend(assay_files)
            return table_files

    def split_assay(self, fp):
        assay_files = []
        header = fp.readline()
        chip_seq_records = [header]
        rna_seq_records = [header]
        me_seq_records = [header]
        tf_seq_records = [header]
        genechip_records = [header]
        chipchip_records = [header]

        default_records = [header]

        assay_types = set()

        is_hybridization_assay = 'hybridization' in get_squashed(header)
        contains_antibody_in_header = 'antibody' in get_squashed(header)

        A = self.ISA.studies[-1].assays[-1]

        log.info(
            "Reading assay memory file; mt=%s, tt=%s",
            A.measurement_type.term,
            A.technology_type.term)
        for line in fp.readlines():
            sqline = get_squashed(line)
            if A.measurement_type and A.technology_type:
                if 'sequencing' in get_squashed(A.technology_type.term) \
                        and 'protein-dnabindingsiteidentification' == \
                        get_squashed(A.measurement_type.term):
                    if not is_hybridization_assay \
                            and 'chip-seq' in sqline \
                            or 'chipseq' in sqline:
                        assay_types.add('ChIP-Seq')
                        chip_seq_records.append(line)
                    if 'bisulfite-seq' in sqline \
                            or 'mre-seq' in sqline \
                            or 'mbd-seq' in sqline \
                            or 'medip-seq' in sqline:
                        assay_types.add('ME-Seq')
                        me_seq_records.append(line)
                    if 'dnase-hypersensitivity' in sqline \
                            or 'mnase-seq' in sqline:
                        assay_types.add('Chromatin-Seq')
                        tf_seq_records.append(line)

            if is_hybridization_assay and (
                    'genomicdna' in sqline
                    or 'genomic_dna' in sqline) \
                    and 'mnase-seq' not in sqline:
                assay_types.add('ChIP-Seq')
                chip_seq_records.append(line)

            if hasattr(A, '_design_type'):
                if 'dye_swap_design' == get_squashed(A._design_type):
                    assay_types.add('Hybridization')
                    genechip_records.append(line)

                if 'chip-chipbytilingarray' in get_squashed(A._design_type):
                    assay_types.add('ChIP-chip by tiling array')
                    chipchip_records.append(line)

            if (is_hybridization_assay and not contains_antibody_in_header) \
                    and 'rna' in sqline \
                    or 'genomicdna' in sqline:
                assay_types.add('transcription profiling by array')
                genechip_records.append(line)

            if (not is_hybridization_assay and ('genomicdna' in sqline
                                                or 'genomic_dna' in sqline)) \
                    and 'mnase-seq' in sqline:
                assay_types.add('ChIP-Seq')
                chip_seq_records.append(line)

            if not is_hybridization_assay and (
                    'rna-seq' in sqline or 'totalrna' in sqline):
                assay_types.add('RNA-Seq')
                rna_seq_records.append(line)

            if is_hybridization_assay and contains_antibody_in_header and (
                    'genomicdna' in sqline or 'chip' in sqline):
                assay_types.add('ChIP-chip')
                chipchip_records.append(line)
            else:
                default_records.append(line)

        log.info("assay_types found: %s", assay_types)

        if len(assay_types) > 0:
            # reset the assays list to load new split ones
            self.ISA.studies[-1].assays = []
            for assay_type in assay_types:
                new_A = copy.copy(A)
                if 'transcription profiling by array' in assay_type:
                    new_A.filename = '{0}-{1}.txt'.format(
                        A.filename[:A.filename.rindex('.')], assay_type)
                    new_A.technology_platform = assay_type
                    a_fp = StringIO()
                    a_fp.writelines(genechip_records)
                    a_fp.name = new_A.filename
                    a_fp.seek(0)
                    self.ISA.studies[-1].assays.append(new_A)
                    assay_files.append(a_fp)
                if 'ChIP-chip' in assay_type:
                    new_A.filename = '{0}-{1}.txt'.format(
                        A.filename[:A.filename.rindex('.')], assay_type)
                    new_A.technology_platform = assay_type
                    a_fp = StringIO()
                    a_fp.writelines(chipchip_records)
                    a_fp.name = new_A.filename
                    a_fp.seek(0)
                    self.ISA.studies[-1].assays.append(new_A)
                    assay_files.append(a_fp)
                if 'ChIP-Seq' in assay_type:
                    new_A.filename = '{0}-{1}.txt'.format(
                        A.filename[:A.filename.rindex('.')], assay_type)
                    new_A.technology_platform = assay_type
                    a_fp = StringIO()
                    a_fp.writelines(chip_seq_records)
                    a_fp.name = new_A.filename
                    a_fp.seek(0)
                    self.ISA.studies[-1].assays.append(new_A)
                    assay_files.append(a_fp)
                if 'RNA-Seq' in assay_type:
                    new_A.filename = '{0}-{1}.txt'.format(
                        A.filename[:A.filename.rindex('.')], assay_type)
                    new_A.technology_platform = assay_type
                    a_fp = StringIO()
                    a_fp.writelines(rna_seq_records)
                    a_fp.name = new_A.filename
                    a_fp.seek(0)
                    self.ISA.studies[-1].assays.append(new_A)
                    assay_files.append(a_fp)
                if 'ME-Seq' in assay_type:
                    new_A.filename = '{0}-{1}.txt'.format(
                        A.filename[:A.filename.rindex('.')], assay_type)
                    new_A.technology_platform = assay_type
                    a_fp = StringIO()
                    a_fp.writelines(me_seq_records)
                    a_fp.name = new_A.filename
                    a_fp.seek(0)
                    self.ISA.studies[-1].assays.append(new_A)
                    assay_files.append(a_fp)
                if 'Chromatin-Seq' in assay_type:
                    new_A.filename = '{0}-{1}.txt'.format(
                        A.filename[:A.filename.rindex('.')], assay_type)
                    new_A.technology_platform = assay_type
                    a_fp = StringIO()
                    a_fp.writelines(tf_seq_records)
                    a_fp.name = new_A.filename
                    a_fp.seek(0)
                    self.ISA.studies[-1].assays.append(new_A)
                    assay_files.append(a_fp)
        else:
            a_fp = StringIO()
            a_fp.writelines(default_records)
            a_fp.name = A.filename
            a_fp.seek(0)
            assay_files.append(a_fp)
        return assay_files


def strip_comments(in_fp):
    """Strip out comment lines indicated by a # at start of line from a given
    file

    :param in_fp: A file-like buffer object
    :return: A memory file buffer object with comments stripped out
    """
    out_fp = StringIO()
    if not isinstance(in_fp, StringIO):
        out_fp.name = in_fp.name
    for line in in_fp.readlines():
        if line.lstrip().startswith('#'):
            pass
        else:
            out_fp.write(line)
    out_fp.seek(0)
    return out_fp
